# Copyright (c) 2023-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Tuple, Optional, Dict, Any

import torch


def get_operator_from_module(module, operator_base_str, dtypes):
    def _get_dtype(dtype):
        if dtype is torch.float32:
            return "fp32"
        if dtype is torch.float16:
            return "fp16"
        if dtype is torch.bfloat16:
            return "bf16"
        if dtype is torch.float64:
            return "fp64"
        if dtype is torch.int16:
            return "int16"
        if dtype is torch.int8:
            return "int8"
        if dtype is torch.int32:
            return "int32"
        if dtype is torch.int64:
            return "int64"

    dtypes = [_get_dtype(dt) for dt in dtypes]
    return getattr(module, operator_base_str + "_" + "_".join(dtypes))


def get_tensor_meta_data(tensor: torch.Tensor) -> Dict[str, Any]:
    if tensor is None:
        return {"size": None, "dtype": None, "device": None}
    return {"size": tensor.size(), "dtype": tensor.dtype, "device": tensor.device}


def maybe_detach(tensor: torch.Tensor) -> Optional[torch.Tensor]:
    return None if tensor is None else tensor.detach().contiguous()


def maybe_size(tensor: torch.Tensor, dim: int = -1) -> int:
    return tensor.size(dim) if tensor is not None else 0


# The following functions are intended for usages where a gradient is only
# initialized if we have the corresponding ctx.needs_input_grads


def maybe_empty_like(
    input: torch.Tensor, create_tensor: bool = True, **kwargs
) -> Optional[torch.Tensor]:
    return torch.empty_like(input, **kwargs) if create_tensor else None


def maybe_empty(
    size: Union[List[int], Tuple[int], torch.Size],
    device: torch.device,
    dtype: torch.dtype,
    create_tensor: bool = True,
) -> Optional[torch.Tensor]:
    return torch.empty(size, device=device, dtype=dtype) if create_tensor else None


def maybe_zeros_like(
    input: torch.Tensor, create_tensor: bool = True
) -> Optional[torch.Tensor]:
    return torch.zeros_like(input) if create_tensor else None


def maybe_zeros(
    size: Union[List[int], Tuple[int], torch.Size],
    device: torch.device,
    dtype: torch.dtype,
    create_tensor: bool = True,
) -> Optional[torch.Tensor]:
    return torch.zeros(size, device=device, dtype=dtype) if create_tensor else None



# Copyright (c) 2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Union

import torch
from torch.autograd.function import once_differentiable

from ..utils import maybe_size, maybe_empty, maybe_detach, get_tensor_meta_data
from ..graph import CSC
from pylibcugraphops.operators import (
    update_efeat_e2e_sum_fwd,
    update_efeat_e2e_sum_bwd,
    update_efeat_e2e_concat_fwd,
    update_efeat_e2e_concat_bwd,
)


def update_efeat_e2e(
    edge_feat: Optional[torch.Tensor],
    node_feat: Union[torch.Tensor, Tuple[Optional[torch.Tensor]]],
    graph: CSC,
    use_source_emb: bool,
    use_target_emb: bool,
    mode: str = "concat",
) -> torch.Tensor:
    """
    PyTorch autograd function for creating new edge features (update_efeat)
    based on either concatenating or summing edge features and the features of
    the corresponding source and destination node of each edge in an edge-to-edge
    fashion (e2e).

    Parameters
    ----------
    edge_feat : torch.Tensor | None
        The input edge features - ignored if passed as ``None``.
        Shape: ``(num_indices, dim_edge)``.

    node_feat : torch.Tensor | Tuple[torch.Tensor] | None
        The input source features. For bipartite graphs, a tuple of
        source and destionations node features with shapes
        ``(num_src_nodes, dim_src)`` and ``(num_dst_nodes, dim_dst)``
        respectively is expected. For non-bipartite graph, a single
        tensor of node features of shape ``(num_dst_nodes, dim_node)``
        is expected with ``dim_src = dim_dst = dim_node``.

    use_source_emb : bool
        whether to use node embeddings indexed by indices of source nodes

    use_target_emb : bool
        whether to use node embeddings indexed by indices of source nodes

    graph : CSC
        The graph used for the operation.

    mode : str, default="concat"
        Mode of update, either ``"concat"`` or ``"sum"``.

    Returns
    -------
    output : torch.Tensor
        The output after either concatenation or summation.
        Shape for ``mode=concat``: ``(num_indices, dim_edge + dim_src + dim_dst)``.
        Shape for ``mode=sum``: ``(num_indices, dim_edge)``.
    """
    modes = ["sum", "concat"]
    if mode.lower() not in modes:
        raise ValueError(f"Supported operations: {modes}, " f"but got '{mode}'.")

    if isinstance(node_feat, Tuple) and len(node_feat) == 2:
        assert (
            graph.is_bipartite
        ), "Expected bipartite graph, but got graph.is_bipartite == False."
        return _update_efeat_e2e_bipartite_autograd.apply(
            edge_feat,
            node_feat[0],
            node_feat[1],
            (graph, mode, use_source_emb, use_target_emb),
        )

    else:
        assert isinstance(node_feat, torch.Tensor) or (node_feat is None)
        assert (
            not graph.is_bipartite
        ), "Expected non-bipartite graph, but got graph.is_bipartite == True."
        return _update_efeat_e2e_autograd.apply(
            edge_feat, node_feat, (graph, mode, use_source_emb, use_target_emb)
        )


class _update_efeat_e2e_autograd(torch.autograd.Function):
    """
    Custom autograd function for an update of the edge features based on
    concatenating or summing edge features, and source node and destination
    node features
    """

    @staticmethod
    def forward(ctx, edge_feat, node_feat, args):
        graph, mode, use_source_emb, use_target_emb = args
        ctx.mode = mode
        ctx.graph = graph

        edge_feat = maybe_detach(edge_feat)
        node_feat = maybe_detach(node_feat)

        dim_edge = maybe_size(edge_feat)
        dim_node = maybe_size(node_feat)

        ctx.dim_edge = dim_edge
        ctx.dim_node = dim_node
        ctx.use_source_emb = use_source_emb
        ctx.use_target_emb = use_target_emb

        if dim_edge > 0:
            dtype = edge_feat.dtype
            device = edge_feat.device

        elif dim_node > 0:
            dtype = node_feat.dtype
            device = node_feat.device

        else:
            raise ValueError("forward called with all the inputs being None")

        stream = torch.cuda.current_stream().cuda_stream

        if mode.lower() == "concat":
            dim_src = dim_node if use_source_emb else 0
            dim_dst = dim_node if use_target_emb else 0
            output = torch.empty(
                graph.num_indices,
                dim_edge + dim_dst + dim_src,
                dtype=dtype,
                device=device,
            )
            update_efeat_e2e_concat_fwd(
                output,
                edge_feat,
                node_feat,
                graph.graph_csc,
                use_source_emb,
                use_target_emb,
                stream,
            )

        else:
            # all should be equal, so find the one which is not 0
            # check for equality is done in binding
            dim_edge = max(dim_edge, dim_node)
            output = torch.empty(
                graph.num_indices, dim_edge, dtype=dtype, device=device
            )
            update_efeat_e2e_sum_fwd(
                output,
                edge_feat,
                node_feat,
                graph.graph_csc,
                use_source_emb,
                use_target_emb,
                stream,
            )

        ctx.node_feat_meta_data = get_tensor_meta_data(node_feat)
        ctx.edge_feat_meta_data = get_tensor_meta_data(edge_feat)

        return output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        grad_output = grad_output.detach().contiguous()
        needs_grad_edge, needs_grad_node, _ = ctx.needs_input_grad

        grad_edge_feat = maybe_empty(
            **ctx.edge_feat_meta_data,
            create_tensor=needs_grad_edge,
        )
        grad_node_feat = maybe_empty(
            **ctx.node_feat_meta_data,
            create_tensor=needs_grad_node,
        )

        stream = torch.cuda.current_stream().cuda_stream
        grad_output = grad_output.detach().contiguous()

        if ctx.mode.lower() == "concat":
            update_efeat_e2e_concat_bwd(
                grad_edge_feat,
                grad_node_feat,
                grad_output,
                ctx.graph.graph_csc,
                ctx.dim_edge,
                ctx.dim_node,
                ctx.use_source_emb,
                ctx.use_target_emb,
                stream,
            )

        else:
            update_efeat_e2e_sum_bwd(
                grad_edge_feat,
                grad_node_feat,
                grad_output,
                ctx.graph.graph_csc,
                ctx.use_source_emb,
                ctx.use_target_emb,
                stream,
            )

        return grad_edge_feat, grad_node_feat, None


class _update_efeat_e2e_bipartite_autograd(torch.autograd.Function):
    """
    Custom autograd function for an update of the edge features based on
    concatenating or summing edge features, and source node and destination
    node features
    """

    @staticmethod
    def forward(ctx, edge_feat, src_node_feat, dst_node_feat, args):
        graph, mode, use_source_emb, use_target_emb = args
        assert (src_node_feat is not None) == use_source_emb
        assert (dst_node_feat is not None) == use_target_emb
        ctx.graph = graph
        ctx.mode = mode

        edge_feat = maybe_detach(edge_feat)
        src_node_feat = maybe_detach(src_node_feat)
        dst_node_feat = maybe_detach(dst_node_feat)

        dim_edge = maybe_size(edge_feat)
        dim_src = maybe_size(src_node_feat)
        dim_dst = maybe_size(dst_node_feat)

        ctx.dim_edge = dim_edge
        ctx.dim_src = dim_src
        ctx.dim_dst = dim_dst

        if dim_edge > 0:
            dtype = edge_feat.dtype
            device = edge_feat.device

        elif dim_src > 0:
            dtype = src_node_feat.dtype
            device = src_node_feat.device

        elif dim_dst > 0:
            dtype = dst_node_feat.dtype
            device = dst_node_feat.device
        else:
            raise ValueError("forward called with all the inputs being None")

        stream = torch.cuda.current_stream().cuda_stream

        if mode.lower() == "concat":
            output = torch.empty(
                graph.num_indices,
                dim_edge + dim_dst + dim_src,
                dtype=dtype,
                device=device,
            )
            update_efeat_e2e_concat_fwd(
                output,
                edge_feat,
                src_node_feat,
                dst_node_feat,
                graph.graph_csc,
                stream,
            )

        else:
            # all should be equal, so find the one which is not 0
            # check for equality is done in binding
            dim_edge = max(dim_edge, dim_src, dim_dst)
            output = torch.empty(
                graph.num_indices, dim_edge, dtype=dtype, device=device
            )
            update_efeat_e2e_sum_fwd(
                output,
                edge_feat,
                src_node_feat,
                dst_node_feat,
                graph.graph_csc,
                stream,
            )

        ctx.src_node_feat_meta_data = get_tensor_meta_data(src_node_feat)
        ctx.dst_node_feat_meta_data = get_tensor_meta_data(dst_node_feat)
        ctx.edge_feat_meta_data = get_tensor_meta_data(edge_feat)

        return output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        grad_output = grad_output.detach().contiguous()
        needs_grad_edge, needs_grad_src, needs_grad_dst, _ = ctx.needs_input_grad

        grad_edge_feat = maybe_empty(
            **ctx.edge_feat_meta_data,
            create_tensor=needs_grad_edge,
        )
        grad_src_feat = maybe_empty(
            **ctx.src_node_feat_meta_data,
            create_tensor=needs_grad_src,
        )
        grad_dst_feat = maybe_empty(
            **ctx.dst_node_feat_meta_data,
            create_tensor=needs_grad_dst,
        )

        stream = torch.cuda.current_stream().cuda_stream
        grad_output = grad_output.detach().contiguous()

        if ctx.mode.lower() == "concat":
            update_efeat_e2e_concat_bwd(
                grad_edge_feat,
                grad_src_feat,
                grad_dst_feat,
                grad_output,
                ctx.graph.graph_csc,
                ctx.dim_edge,
                ctx.dim_src,
                ctx.dim_dst,
                stream,
            )

        else:
            update_efeat_e2e_sum_bwd(
                grad_edge_feat,
                grad_src_feat,
                grad_dst_feat,
                grad_output,
                ctx.graph.graph_csc,
                stream,
            )

        return grad_edge_feat, grad_src_feat, grad_dst_feat, None



# Copyright (c) 2023-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from typing import Optional, Union, Tuple
from torch.autograd.function import once_differentiable

from ..utils import (
    maybe_empty_like,
    maybe_detach,
    maybe_empty,
    maybe_zeros_like,
    get_operator_from_module,
)
from ..graph import CSC
from pylibcugraphops import ActivationOp, mha_params
from pylibcugraphops.operators import activation_params
import pylibcugraphops.operators as ops

_activation_ops = [x for x in dir(ActivationOp) if not x.startswith("__")]


def mha_gat_n2n(
    feat: Union[torch.Tensor, Tuple[torch.Tensor]],
    attn_weights: torch.Tensor,
    graph: CSC,
    num_heads: int = 1,
    activation: str = "LeakyReLU",
    negative_slope: float = 0.2,
    concat_heads: bool = True,
    edge_feat: Optional[torch.Tensor] = None,
    deterministic_dgrad: bool = False,
    deterministic_wgrad: bool = False,
    high_precision_dgrad: bool = False,
    high_precision_wgrad: bool = False,
) -> torch.Tensor:
    """
    PyTorch autograd function for a multi-head attention layer (GAT-like)
    without using cudnn (mha_gat) in a node-to-node reduction (n2n).

    Parameters
    ----------
    feat : torch.Tensor | Tuple[torch.Tensor, torch.Tensor]
        The input node features.
        Depending on whether the graph is bipartite or not, this has
        to be a tuple of two tensors or a single tensor correspondingly.
        For bipartite graphs, the rows of each tensors consists of
        concatenated features from different heads after the linear transformation.
        Shapes: ``(n_src_nodes, dim_in), (n_dst_nodes, dim_in)``,
        where ``dim_in = dim_head * num_heads``, with ``dim_head`` being the
        feature dimension per head.
        For non-bipartite graph, the rows of the single tensor consists of
        concatenated features from different heads after the linear transformation.
        Shape: ``(n_src_nodes, dim_in)``, where ``dim_in = dim_head * num_heads``,
        with ``dim_head`` being the feature dimension per head.

    attn_weights : torch.Tensor
        The attention weights with ``dim_w = 2 * dim_in + dim_in_edges`` where
        ``dim_in_edges`` is assumed to be 0 if ``edge_feat`` is None.
        Shape: ``(dim_w,)``.

    graph : CSC
        The graph used in for the operation.

    num_heads : int, default=1
        Number of heads in multi-head attention.

    activation : str, default="LeakyReLU"
        The activation function used in the attention mechanism. Choose from
        ``"ELU"``, ``"LeakyReLU"``, ``"Linear"``, ``"ReLU"``, ``"Scalar"``,
        ``"Sigmoid"`` and ``"Tanh"``.

    negative_slope : float, default=0.2
        LeakyReLU angle of negative slope. Only effective when ``activation`` is
        ``"LeakyReLU"``.

    concat_heads : bool, default=True
        Aggregated embeddings from each head are concatenated if ``True`` or
        averaged if ``False``.

    edge_feat : Optional[torch.Tensor], default=None
        Optional input edge features. Each row consists of concatenated features from
        different heads after the linear transformation.
        Shape: ``(n_edges, dim_in_edge)`` where
        ``dim_in_edge = dim_head_edge * num_heads``, with ``dim_head_edge``
        being the feature dimension per head.

    deterministic_dgrad : Optional[bool], default=False
        Optional flag indicating whether the feature gradients
        are computed deterministically using a dedicated workspace buffer.

    deterministic_wgrad : Optional[bool], default=False
        Optional flag indicating whether the weight gradients
        are computed deterministically using a dedicated workspace buffer.

    high_precision_dgrad : Optional[bool], default=False
        Optional flag indicating whether gradients for inputs in half precision
        are kept in single precision as long as possible and only casted to
        the corresponding input type at the very end.

    high_precision_wgrad : Optional[bool], default=False
        Optional flag indicating whether gradients for weights in half precision
        are kept in single precision as long as possible and only casted to
        the corresponding input type at the very end.

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape for ``concat_heads=True``: ``(n_dst_nodes, dim_in)``;
        Shape for ``concat_heads=False``: ``(n_dst_nodes, dim_head)``.
    """
    if activation not in _activation_ops:
        raise ValueError(
            f"Supported activation functions: {_activation_ops}, "
            f"but got '{activation}'."
        )

    act_params = activation_params(negative_slope, getattr(ActivationOp, activation))
    params = mha_params(act_params, num_heads, concat_heads)
    if graph.is_bipartite:
        assert isinstance(feat, Tuple) and len(feat) == 2, (
            f"Expected tuple of source and destination node features,"
            f" got {type(feat)} instead."
        )
        return _mha_gat_n2n_bipartite_autograd.apply(
            feat[0],
            feat[1],
            edge_feat,
            attn_weights,
            graph,
            params,
            deterministic_dgrad,
            deterministic_wgrad,
            high_precision_dgrad,
            high_precision_wgrad,
        )

    assert isinstance(
        feat, torch.Tensor
    ), f"Expected torch.Tensor of node features, got {type(feat)} instead."
    return _mha_gat_n2n_autograd.apply(
        feat,
        edge_feat,
        attn_weights,
        graph,
        params,
        deterministic_dgrad,
        deterministic_wgrad,
        high_precision_dgrad,
        high_precision_wgrad,
    )


class _mha_gat_n2n_autograd(torch.autograd.Function):
    """
    Custom autograd function for a multi-head attention layer (GAT-like)
    without using cudnn (mha_gat) operating on a graph in a node-to-node
    reduction (n2n).
    """

    @staticmethod
    def forward(
        ctx,
        feat,
        edge_feat,
        attn_weights,
        graph,
        params,
        deterministic_dgrad,
        deterministic_wgrad,
        high_precision_dgrad,
        high_precision_wgrad,
    ):
        dim = feat.size(-1)
        dim_out = dim if params.concat_heads else (dim // params.num_heads)
        output = torch.empty(
            graph.num_dst_nodes, dim_out, dtype=feat.dtype, device=feat.device
        )

        feat = feat.detach().contiguous()
        attn_weights = attn_weights.detach().contiguous()
        edge_feat = maybe_detach(edge_feat)

        sm_scores = torch.empty(
            2,
            params.num_heads,
            graph.num_indices,
            dtype=torch.float32,
            device=feat.device,
        )

        stream = torch.cuda.current_stream().cuda_stream

        if graph.num_dst_nodes > 0:
            if edge_feat is None:
                fwd_fun = get_operator_from_module(
                    ops,
                    "mha_gat_n2n_fwd",
                    (
                        output.dtype,
                        sm_scores.dtype,
                    ),
                )
                fwd_fun(
                    output,
                    sm_scores,
                    feat,
                    attn_weights,
                    graph.graph_csc,
                    params,
                    stream_id=stream,
                )

            else:
                fwd_fun = get_operator_from_module(
                    ops,
                    "mha_gat_n2n_efeat_fwd",
                    (
                        output.dtype,
                        sm_scores.dtype,
                    ),
                )

                fwd_fun(
                    output,
                    sm_scores,
                    feat,
                    edge_feat,
                    attn_weights,
                    graph.graph_csc,
                    params,
                    stream_id=stream,
                )

        ctx.graph = graph
        ctx.params = params
        ctx.deterministic_dgrad = deterministic_dgrad
        ctx.deterministic_wgrad = deterministic_wgrad
        ctx.high_precision_dgrad = high_precision_dgrad
        ctx.high_precision_wgrad = high_precision_wgrad
        ctx.save_for_backward(feat, edge_feat, attn_weights, sm_scores)

        return output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        feat, edge_feat, attn_weights, sm_scores = ctx.saved_tensors
        grad_output = grad_output.detach().contiguous()
        (
            needs_grad_feat,
            needs_grad_efeat,
            needs_grad_weights,
            _,
            _,
            _,
            _,
            _,
            _,
        ) = ctx.needs_input_grad
        if not (needs_grad_feat or needs_grad_weights or needs_grad_efeat):
            return None, None, None, None, None, None, None, None, None

        if ctx.graph.num_dst_nodes > 0:
            dgrad_dtype = (
                torch.float32 if ctx.high_precision_dgrad else grad_output.dtype
            )
            wgrad_dtype = (
                torch.float32 if ctx.high_precision_dgrad else grad_output.dtype
            )
            grad_feat = maybe_empty_like(feat, needs_grad_feat, dtype=dgrad_dtype)
            grad_edge_feat = maybe_empty_like(
                edge_feat, needs_grad_efeat, dtype=dgrad_dtype
            )
            grad_attn_weights = maybe_empty_like(
                attn_weights, needs_grad_weights, dtype=wgrad_dtype
            )
            grad_wspace_feat = maybe_empty(
                (ctx.graph.num_indices, feat.size(1)),
                create_tensor=(needs_grad_feat and ctx.deterministic_dgrad),
                dtype=dgrad_dtype,
                device=attn_weights.device,
            )
            grad_wspace_weight = maybe_empty(
                (ctx.graph.num_dst_nodes, attn_weights.size(0)),
                create_tensor=(needs_grad_weights and ctx.deterministic_wgrad),
                dtype=wgrad_dtype,
                device=attn_weights.device,
            )
            grad_sm_scores = torch.empty_like(sm_scores)

            stream = torch.cuda.current_stream().cuda_stream

            if edge_feat is None:
                bwd_fun = get_operator_from_module(
                    ops,
                    "mha_gat_n2n_bwd",
                    (
                        grad_output.dtype,
                        dgrad_dtype,
                        wgrad_dtype,
                        sm_scores.dtype,
                    ),
                )
                bwd_fun(
                    grad_feat,
                    grad_attn_weights if grad_wspace_weight is None else None,
                    grad_sm_scores,
                    grad_output,
                    feat,
                    attn_weights,
                    sm_scores,
                    ctx.graph.graph_csc,
                    ctx.params,
                    grad_wspace_feat,
                    grad_wspace_weight,
                    stream_id=stream,
                )

                if (grad_wspace_weight is not None) and (grad_attn_weights is not None):
                    grad_attn_weights = grad_wspace_weight.sum(dim=0)

            else:
                bwd_fun = get_operator_from_module(
                    ops,
                    "mha_gat_n2n_efeat_bwd",
                    (
                        grad_output.dtype,
                        dgrad_dtype,
                        wgrad_dtype,
                        sm_scores.dtype,
                    ),
                )
                bwd_fun(
                    grad_feat,
                    grad_edge_feat,
                    grad_attn_weights if grad_wspace_weight is None else None,
                    grad_sm_scores,
                    grad_output,
                    feat,
                    edge_feat,
                    attn_weights,
                    sm_scores,
                    ctx.graph.graph_csc,
                    ctx.params,
                    grad_wspace_feat,
                    grad_wspace_weight,
                    stream_id=stream,
                )

                if (grad_wspace_weight is not None) and (grad_attn_weights is not None):
                    grad_attn_weights = grad_wspace_weight.sum(dim=0)

            if (grad_feat is not None) and (grad_feat.dtype != feat.dtype):
                grad_feat = grad_feat.to(dtype=feat.dtype)

            if (grad_edge_feat is not None) and (
                grad_edge_feat.dtype != edge_feat.dtype
            ):
                grad_edge_feat = grad_edge_feat.to(dtype=edge_feat.dtype)

            if (grad_attn_weights is not None) and (
                grad_attn_weights.dtype != attn_weights.dtype
            ):
                grad_attn_weights = grad_attn_weights.to(dtype=attn_weights.dtype)

        else:
            grad_feat = maybe_zeros_like(feat, needs_grad_feat)
            grad_edge_feat = maybe_zeros_like(edge_feat, needs_grad_efeat)
            grad_attn_weights = maybe_zeros_like(attn_weights, needs_grad_weights)

        return (
            grad_feat,
            grad_edge_feat,
            grad_attn_weights,
            None,
            None,
            None,
            None,
            None,
            None,
        )


class _mha_gat_n2n_bipartite_autograd(torch.autograd.Function):
    """
    Custom autograd function for a multi-head attention layer (GAT-like)
    without using cudnn (mha_gat) operating on bipartite graphs in a node-to-node
    reduction (n2n).
    """

    @staticmethod
    def forward(
        ctx,
        src_feat,
        dst_feat,
        edge_feat,
        attn_weights,
        graph,
        params,
        deterministic_dgrad,
        deterministic_wgrad,
        high_precision_dgrad,
        high_precision_wgrad,
    ):
        dim = src_feat.size(-1)
        dim_out = dim if params.concat_heads else (dim // params.num_heads)
        output = torch.empty(
            graph.num_dst_nodes, dim_out, dtype=src_feat.dtype, device=src_feat.device
        )
        src_feat = src_feat.detach().contiguous()
        dst_feat = dst_feat.detach().contiguous()
        attn_weights = attn_weights.detach().contiguous()
        edge_feat = maybe_detach(edge_feat)

        sm_scores = torch.empty(
            2,
            params.num_heads,
            graph.num_indices,
            dtype=torch.float32,
            device=src_feat.device,
        )

        stream = torch.cuda.current_stream().cuda_stream

        if graph.num_dst_nodes > 0:
            if edge_feat is None:
                fwd_fun = get_operator_from_module(
                    ops,
                    "mha_gat_n2n_fwd",
                    (
                        output.dtype,
                        sm_scores.dtype,
                    ),
                )
                fwd_fun(
                    output,
                    sm_scores,
                    src_feat,
                    dst_feat,
                    attn_weights,
                    graph.graph_csc,
                    params,
                    stream_id=stream,
                )

            else:
                fwd_fun = get_operator_from_module(
                    ops,
                    "mha_gat_n2n_efeat_fwd",
                    (
                        output.dtype,
                        sm_scores.dtype,
                    ),
                )
                fwd_fun(
                    output,
                    sm_scores,
                    src_feat,
                    dst_feat,
                    edge_feat,
                    attn_weights,
                    graph.graph_csc,
                    params,
                    stream_id=stream,
                )

        ctx.graph = graph
        ctx.params = params
        ctx.deterministic_dgrad = deterministic_dgrad
        ctx.deterministic_wgrad = deterministic_wgrad
        ctx.high_precision_dgrad = high_precision_dgrad
        ctx.high_precision_wgrad = high_precision_wgrad
        ctx.save_for_backward(src_feat, dst_feat, edge_feat, attn_weights, sm_scores)

        return output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        src_feat, dst_feat, edge_feat, attn_weights, sm_scores = ctx.saved_tensors
        grad_output = grad_output.detach().contiguous()
        (
            needs_grad_src_feat,
            needs_grad_dst_feat,
            needs_grad_efeat,
            needs_grad_weights,
            _,
            _,
            _,
            _,
            _,
            _,
        ) = ctx.needs_input_grad
        if not (
            needs_grad_src_feat
            or needs_grad_dst_feat
            or needs_grad_weights
            or needs_grad_efeat
        ):
            return None, None, None, None, None, None, None, None, None

        if ctx.graph.num_dst_nodes > 0:
            dgrad_dtype = (
                torch.float32 if ctx.high_precision_dgrad else grad_output.dtype
            )
            wgrad_dtype = (
                torch.float32 if ctx.high_precision_wgrad else grad_output.dtype
            )
            grad_src_feat = maybe_empty_like(
                src_feat, needs_grad_src_feat, dtype=dgrad_dtype
            )
            grad_dst_feat = maybe_empty_like(
                dst_feat, needs_grad_dst_feat, dtype=dgrad_dtype
            )
            grad_edge_feat = maybe_empty_like(
                edge_feat, needs_grad_efeat, dtype=dgrad_dtype
            )
            grad_attn_weights = maybe_empty_like(
                attn_weights, needs_grad_weights, dtype=wgrad_dtype
            )
            grad_wspace_feat = maybe_empty(
                (ctx.graph.num_indices, src_feat.size(1)),
                create_tensor=(needs_grad_dst_feat and ctx.deterministic_dgrad),
                dtype=dgrad_dtype,
                device=attn_weights.device,
            )
            grad_wspace_weight = maybe_empty(
                (ctx.graph.num_dst_nodes, attn_weights.size(0)),
                create_tensor=(needs_grad_weights and ctx.deterministic_wgrad),
                dtype=wgrad_dtype,
                device=attn_weights.device,
            )
            grad_sm_scores = torch.empty_like(sm_scores)

            stream = torch.cuda.current_stream().cuda_stream

            if edge_feat is None:
                bwd_fun = get_operator_from_module(
                    ops,
                    "mha_gat_n2n_bwd",
                    (
                        grad_output.dtype,
                        dgrad_dtype,
                        wgrad_dtype,
                        sm_scores.dtype,
                    ),
                )
                bwd_fun(
                    grad_src_feat,
                    grad_dst_feat,
                    grad_attn_weights if grad_wspace_weight is None else None,
                    grad_sm_scores,
                    grad_output,
                    src_feat,
                    dst_feat,
                    attn_weights,
                    sm_scores,
                    ctx.graph.graph_csc,
                    ctx.params,
                    grad_wspace_feat,
                    grad_wspace_weight,
                    stream_id=stream,
                )

                if (grad_wspace_weight is not None) and (grad_attn_weights is not None):
                    grad_attn_weights = grad_wspace_weight.sum(dim=0)

            else:
                bwd_fun = get_operator_from_module(
                    ops,
                    "mha_gat_n2n_efeat_bwd",
                    (
                        grad_output.dtype,
                        dgrad_dtype,
                        wgrad_dtype,
                        sm_scores.dtype,
                    ),
                )
                bwd_fun(
                    grad_src_feat,
                    grad_dst_feat,
                    grad_edge_feat,
                    grad_attn_weights if grad_wspace_weight is None else None,
                    grad_sm_scores,
                    grad_output,
                    src_feat,
                    dst_feat,
                    edge_feat,
                    attn_weights,
                    sm_scores,
                    ctx.graph.graph_csc,
                    ctx.params,
                    grad_wspace_feat,
                    grad_wspace_weight,
                    stream_id=stream,
                )

                if (grad_wspace_weight is not None) and (grad_attn_weights is not None):
                    grad_attn_weights = grad_wspace_weight.sum(dim=0)

            if (grad_src_feat is not None) and (grad_src_feat.dtype != src_feat.dtype):
                grad_src_feat = grad_src_feat.to(dtype=src_feat.dtype)

            if (grad_dst_feat is not None) and (grad_dst_feat.dtype != dst_feat.dtype):
                grad_dst_feat = grad_dst_feat.to(dtype=dst_feat.dtype)

            if (grad_edge_feat is not None) and (
                grad_edge_feat.dtype != edge_feat.dtype
            ):
                grad_edge_feat = grad_edge_feat.to(dtype=edge_feat.dtype)

            if (grad_attn_weights is not None) and (
                grad_attn_weights.dtype != attn_weights.dtype
            ):
                grad_attn_weights = grad_attn_weights.to(dtype=attn_weights.dtype)

        else:
            grad_src_feat = maybe_zeros_like(src_feat, needs_grad_src_feat)
            grad_dst_feat = maybe_zeros_like(dst_feat, needs_grad_dst_feat)
            grad_edge_feat = maybe_zeros_like(edge_feat, needs_grad_efeat)
            grad_attn_weights = maybe_zeros_like(attn_weights, needs_grad_weights)

        return (
            grad_src_feat,
            grad_dst_feat,
            grad_edge_feat,
            grad_attn_weights,
            None,
            None,
            None,
            None,
            None,
            None,
        )



# Copyright (c) 2023-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from typing import Optional, Tuple, Union
from torch.autograd.function import once_differentiable

from ..utils import (
    maybe_empty_like,
    maybe_empty,
    get_tensor_meta_data,
    maybe_zeros,
    maybe_zeros_like,
)
from ..graph import CSC
from pylibcugraphops.operators import (
    AggOp,
    agg_concat_n2n_fwd,
    agg_concat_n2n_bwd,
    agg_concat_e2n_fwd,
    agg_concat_e2n_bwd,
    agg_concat_n2n_e2n_fwd,
    agg_concat_n2n_e2n_bwd,
    agg_concat_weighted_n2n_fwd,
    agg_concat_weighted_n2n_bwd,
    agg_concat_n2n_bipartite_fwd,
    agg_concat_n2n_bipartite_bwd,
    agg_concat_e2n_bipartite_fwd,
    agg_concat_e2n_bipartite_bwd,
    agg_concat_n2n_e2n_bipartite_fwd,
    agg_concat_n2n_e2n_bipartite_bwd,
    agg_concat_weighted_n2n_bipartite_fwd,
    agg_concat_weighted_n2n_bipartite_bwd,
)

_aggr_ops = ["max", "mean", "min", "sum"]


def agg_concat_n2n(
    feat: Union[torch.Tensor, Tuple[torch.Tensor]],
    graph: CSC,
    aggr: str = "sum",
    edge_weight: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """
    PyTorch autograd function for simple aggregation using node features in
    an node-to-node reduction (n2n) while concatenating the original features of
    output at the end (agg_concat).

    Parameters
    ----------
    feat : torch.Tensor | Tuple[torch.Tensor]
        The input node features.
        Shape: ``(n_src_nodes, dim_in)`` if graph is non-bipartite.
        Shape: ``(n_src_nodes, dim_in)`` and ``(n_dst_nodes, dim_concat)``
        if graph is bipartite.

    graph : CSC
        The graph used for the operation.

    aggr : str, default="sum"
        Aggregation operation. Choose from ``"max"``, ``"mean"``, ``"min"``, ``"sum"``.

    edge_weight : Optional[torch.Tensor], default=None
        When passing additional edge weights, the aggregation is done in
        a weighted fashion. Note that for min/max the result after
        weighting features with those edge weights is returned.
        Edge weights are indexed in the same fashion as the indices in ``graph``.
        Shape: ``(n_edges, )``.

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape: ``(n_dst_nodes, 2 * dim_in)``.
    """

    if aggr not in _aggr_ops:
        raise ValueError(
            f"Supported aggregation operations: {_aggr_ops}, " f"but got '{aggr}'."
        )
    aggr = getattr(AggOp, aggr.capitalize())

    if graph.is_bipartite:
        assert isinstance(feat, Tuple) and len(feat) == 2, (
            f"Expected tuple of source and destination node features,"
            f" got {type(feat)} instead."
        )
        return _agg_concat_n2n_bipartite_autograd.apply(
            feat[0], feat[1], edge_weight, graph, aggr
        )

    assert isinstance(
        feat, torch.Tensor
    ), f"Expected torch.Tensor of node features, got {type(feat)} instead."
    return _agg_concat_n2n_autograd.apply(feat, edge_weight, graph, aggr)


def agg_concat_e2n(
    node_feat: torch.Tensor,
    edge_feat: torch.Tensor,
    graph: CSC,
    aggr: str = "sum",
) -> torch.Tensor:
    r"""PyTorch autograd function for simple aggregation using edge features in
    while concatenating the original features of output at the end (agg_concat).

    Parameters
    ----------
    node_feat : torch.Tensor
        The input node features.
        Shape: ``(n_dst_nodes, dim_node)``.

    edge_feat : torch.Tensor
        The input edge features.
        Shape: ``(n_edges, dim_edge)``.

    graph : CSC
        The graph used for the operation.

    aggr : str, default="sum"
        Aggregation operation. Choose from ``"max"``, ``"mean"``, ``"min"``, ``"sum"``.

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape: ``(n_dst_nodes, dim_node + dim_edge)``.
    """

    if aggr not in _aggr_ops:
        raise ValueError(
            f"Supported aggregation operations: {_aggr_ops}, " f"but got '{aggr}'."
        )
    aggr = getattr(AggOp, aggr.capitalize())

    return _agg_concat_e2n_autograd.apply(edge_feat, node_feat, graph, aggr)


def agg_concat_n2n_e2n(
    feat: Union[torch.Tensor, Tuple[torch.Tensor]],
    edge_feat: torch.Tensor,
    graph: CSC,
    aggr: str = "sum",
) -> torch.Tensor:
    r"""PyTorch autograd function for simple aggregation using node features in
    an node-to-node reduction (n2n) and edge features in and edge-to-node
    reduction (e2n) while concatenating the original features of output at the
    end (agg_concat).

    Parameters
    ----------
    node_feat : torch.Tensor
        The input node features.
        Shape: ``(n_src_nodes, dim_in)`` if graph is non-bipartite.
        Shape: ``(n_src_nodes, dim_in)`` and ``(n_dst_nodes, dim_concat)``
        if graph is bipartite.

    edge_feat : torch.Tensor
        The input edge features.
        Shape: ``(n_edges, dim_edge)``.

    graph : CSC
        The graph used for the operation.

    aggr : str, default="sum"
        Aggregation operation. Choose from ``"max"``, ``"mean"``, ``"min"``, ``"sum"``.

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape: ``(n_dst_nodes, dim_node + dim_edge + dim_node)``.
    """

    if aggr not in _aggr_ops:
        raise ValueError(
            f"Supported aggregation operations: {_aggr_ops}, " f"but got '{aggr}'."
        )
    aggr = getattr(AggOp, aggr.capitalize())

    if graph.is_bipartite:
        assert isinstance(feat, Tuple) and len(feat) == 2, (
            f"Expected tuple of source and destination node features,"
            f" got {type(feat)} instead."
        )
        return _agg_concat_n2n_e2n_bipartite_autograd.apply(
            feat[0],
            edge_feat,
            feat[1],
            graph,
            aggr,
        )

    assert isinstance(
        feat, torch.Tensor
    ), f"Expected torch.Tensor of node features, got {type(feat)} instead."
    return _agg_concat_n2n_e2n_autograd.apply(feat, edge_feat, graph, aggr)


class _agg_concat_n2n_autograd(torch.autograd.Function):
    """
    Custom autograd function for simple aggregation using node features in
    an node-to-node reduction (n2n) on a graph while concatenating the
    original features of output at the end (agg_concat).
    """

    @staticmethod
    def forward(ctx, feat, edge_weight, graph, aggr):
        dim_in = feat.size(-1)
        agg_output = torch.empty(
            graph.num_dst_nodes, 2 * dim_in, dtype=feat.dtype, device=feat.device
        )

        node_deg = None
        out_pos = None
        if graph.num_dst_nodes > 0:
            if aggr in (AggOp.Max, AggOp.Min):
                idx_type = graph.graph_csc.__class__.__name__.split("_")[-1]
                out_pos = torch.empty(
                    graph.num_dst_nodes,
                    dim_in,
                    dtype=getattr(torch, idx_type),
                    device=feat.device,
                )

            stream = torch.cuda.current_stream().cuda_stream
            feat = feat.detach().contiguous()

            if edge_weight is None:
                agg_concat_n2n_fwd(
                    agg_output, feat, graph.graph_csc, aggr, out_pos, stream_id=stream
                )

            else:
                edge_weight = edge_weight.detach().contiguous()
                node_deg = torch.empty(
                    graph.num_dst_nodes,
                    device=edge_weight.device,
                    dtype=edge_weight.dtype,
                )

                agg_concat_weighted_n2n_fwd(
                    agg_output,
                    feat,
                    edge_weight,
                    graph.graph_csc,
                    aggr,
                    out_pos,
                    node_deg,
                    stream_id=stream,
                )

        ctx.graph = graph
        ctx.aggr = aggr
        ctx.dim_node = dim_in
        ctx.feat_meta_data = get_tensor_meta_data(feat)

        saved_agg_output = (
            agg_output.detach().contiguous()
            if (edge_weight is not None) and (aggr == AggOp.Mean)
            else None
        )
        saved_feat = feat if edge_weight is not None else None
        ctx.save_for_backward(
            saved_feat, edge_weight, node_deg, out_pos, saved_agg_output
        )

        return agg_output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        feat, edge_weight, node_deg, out_pos, agg_output = ctx.saved_tensors
        needs_grad_feat, needs_grad_weight, _, _ = ctx.needs_input_grad
        if not (needs_grad_feat or needs_grad_weight):
            return None, None, None, None

        grad_output = grad_output.detach().contiguous()

        if ctx.graph.num_dst_nodes > 0:
            grad_feat = maybe_empty(
                **ctx.feat_meta_data,
                create_tensor=needs_grad_feat,
            )
            # captures both having edge_weight=None and
            # needs_grad_weight=False with edge_weight!=None
            grad_edge_weight = maybe_empty_like(edge_weight, needs_grad_weight)
            stream = torch.cuda.current_stream().cuda_stream

            if edge_weight is None:
                agg_concat_n2n_bwd(
                    grad_feat,
                    grad_output,
                    ctx.graph.graph_csc,
                    ctx.dim_node,
                    ctx.aggr,
                    out_pos,
                    stream_id=stream,
                )

            else:
                agg_concat_weighted_n2n_bwd(
                    grad_feat,
                    grad_edge_weight,
                    grad_output,
                    edge_weight,
                    feat,
                    ctx.graph.graph_csc,
                    ctx.dim_node,
                    ctx.aggr,
                    agg_output,
                    out_pos,
                    node_deg,
                    stream_id=stream,
                )
        else:
            grad_feat = maybe_zeros(
                **ctx.feat_meta_data,
                create_tensor=needs_grad_feat,
            )
            grad_edge_weight = maybe_zeros_like(edge_weight, needs_grad_weight)

        return grad_feat, grad_edge_weight, None, None


class _agg_concat_n2n_bipartite_autograd(torch.autograd.Function):
    """
    Custom autograd function for simple aggregation using node features in
    an node-to-node reduction (n2n) on a graph while concatenating the
    original features of output at the end (agg_concat).
    """

    @staticmethod
    def forward(ctx, src_feat, concat_feat, edge_weight, graph, aggr):
        dim_in = src_feat.size(-1)
        dim_concat = concat_feat.size(-1)

        agg_output = torch.empty(
            graph.num_dst_nodes,
            dim_in + dim_concat,
            dtype=src_feat.dtype,
            device=src_feat.device,
        )
        node_deg = None
        out_pos = None
        if graph.num_dst_nodes > 0:
            if aggr in (AggOp.Max, AggOp.Min):
                idx_type = graph.graph_csc.__class__.__name__.split("_")[-1]
                out_pos = torch.empty(
                    graph.num_dst_nodes,
                    dim_in,
                    dtype=getattr(torch, idx_type),
                    device=src_feat.device,
                )

            stream = torch.cuda.current_stream().cuda_stream

            src_feat = src_feat.detach().contiguous()
            concat_feat = concat_feat.detach().contiguous()
            if edge_weight is None:
                agg_concat_n2n_bipartite_fwd(
                    agg_output,
                    src_feat,
                    concat_feat,
                    graph.graph_csc,
                    aggr,
                    out_pos,
                    stream_id=stream,
                )

            else:
                edge_weight = edge_weight.detach().contiguous()
                node_deg = torch.empty(
                    graph.num_dst_nodes,
                    device=edge_weight.device,
                    dtype=edge_weight.dtype,
                )

                agg_concat_weighted_n2n_bipartite_fwd(
                    agg_output,
                    src_feat,
                    concat_feat,
                    edge_weight,
                    graph.graph_csc,
                    aggr,
                    out_pos,
                    node_deg,
                    stream_id=stream,
                )

        ctx.graph = graph
        ctx.aggr = aggr
        ctx.dim_node = dim_in
        ctx.dim_concat = dim_concat
        ctx.feat_meta_data = get_tensor_meta_data(src_feat)
        ctx.concat_meta_data = get_tensor_meta_data(concat_feat)

        saved_agg_output = (
            agg_output.detach().contiguous()
            if (edge_weight is not None) and (aggr == AggOp.Mean)
            else None
        )
        saved_feat = src_feat if edge_weight is not None else None
        ctx.save_for_backward(
            saved_feat, edge_weight, node_deg, out_pos, saved_agg_output
        )

        return agg_output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        feat, edge_weight, node_deg, out_pos, agg_output = ctx.saved_tensors
        (
            needs_grad_feat,
            needs_grad_concat,
            needs_grad_weight,
            _,
            _,
        ) = ctx.needs_input_grad
        if not (needs_grad_feat or needs_grad_weight or needs_grad_concat):
            return None, None, None, None, None

        grad_output = grad_output.detach().contiguous()

        if ctx.graph.num_dst_nodes > 0:
            grad_feat = maybe_empty(
                **ctx.feat_meta_data,
                create_tensor=needs_grad_feat,
            )
            grad_concat = maybe_empty(
                **ctx.concat_meta_data,
                create_tensor=needs_grad_concat,
            )

            # captures both having edge_weight=None and
            # needs_grad_weight=False with edge_weight!=None
            grad_edge_weight = maybe_empty_like(edge_weight, needs_grad_weight)
            stream = torch.cuda.current_stream().cuda_stream

            if edge_weight is None:
                agg_concat_n2n_bipartite_bwd(
                    grad_feat,
                    grad_concat,
                    grad_output,
                    ctx.graph.graph_csc,
                    ctx.dim_node,
                    ctx.dim_concat,
                    ctx.aggr,
                    out_pos,
                    stream_id=stream,
                )

            else:
                agg_concat_weighted_n2n_bipartite_bwd(
                    grad_feat,
                    grad_concat,
                    grad_edge_weight,
                    grad_output,
                    edge_weight,
                    feat,
                    ctx.graph.graph_csc,
                    ctx.dim_node,
                    ctx.dim_concat,
                    ctx.aggr,
                    agg_output,
                    out_pos,
                    node_deg,
                    stream_id=stream,
                )
        else:
            grad_feat = maybe_zeros(
                **ctx.feat_meta_data,
                create_tensor=needs_grad_feat,
            )
            grad_concat = maybe_zeros(
                **ctx.concat_meta_data,
                create_tensor=needs_grad_concat,
            )
            grad_edge_weight = maybe_zeros_like(edge_weight, needs_grad_weight)

        return grad_feat, grad_concat, grad_edge_weight, None, None


class _agg_concat_e2n_autograd(torch.autograd.Function):
    r"""Custom autograd function for simple aggregation using edge features in
    an edge-to-node reduction (e2n) on a graph wile concatenating the
    original features of output at the end (agg_concat)."""

    @staticmethod
    def forward(ctx, edge_feat, node_feat, graph, aggr):
        dim_node = node_feat.size(-1)
        dim_edge = edge_feat.size(-1)
        agg_output = torch.empty(
            graph.num_dst_nodes,
            dim_edge + dim_node,
            dtype=edge_feat.dtype,
            device=edge_feat.device,
        )

        out_pos = None
        if graph.num_dst_nodes > 0:
            if aggr in (AggOp.Max, AggOp.Min):
                idx_type = graph.graph_csc.__class__.__name__.split("_")[-1]
                out_pos = torch.empty(
                    graph.num_dst_nodes,
                    dim_edge,
                    dtype=getattr(torch, idx_type),
                    device=edge_feat.device,
                )

            stream = torch.cuda.current_stream().cuda_stream
            node_feat = node_feat.detach().contiguous()
            edge_feat = edge_feat.detach().contiguous()
            if graph.is_bipartite:
                agg_concat_e2n_bipartite_fwd(
                    agg_output,
                    edge_feat,
                    node_feat,
                    graph.graph_csc,
                    aggr,
                    out_pos,
                    stream_id=stream,
                )
            else:
                agg_concat_e2n_fwd(
                    agg_output,
                    edge_feat,
                    node_feat,
                    graph.graph_csc,
                    aggr,
                    out_pos,
                    stream_id=stream,
                )

        ctx.graph = graph
        ctx.aggr = aggr
        ctx.dim_node = dim_node
        ctx.dim_edge = dim_edge

        ctx.edge_feat_meta_data = get_tensor_meta_data(edge_feat)
        ctx.node_feat_meta_data = get_tensor_meta_data(node_feat)

        ctx.save_for_backward(out_pos)

        return agg_output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        grad_output = grad_output.detach().contiguous()
        (out_pos,) = ctx.saved_tensors
        needs_grad_edge, needs_grad_node, _, _ = ctx.needs_input_grad
        if not (needs_grad_node or needs_grad_edge):
            return None, None, None, None

        grad_output = grad_output.detach().contiguous()

        if ctx.graph.num_dst_nodes > 0:
            grad_node_feat = maybe_empty(
                **ctx.node_feat_meta_data,
                create_tensor=needs_grad_node,
            )
            grad_edge_feat = maybe_empty(
                **ctx.edge_feat_meta_data,
                create_tensor=needs_grad_edge,
            )

            stream = torch.cuda.current_stream().cuda_stream

            if ctx.graph.is_bipartite:
                agg_concat_e2n_bipartite_bwd(
                    grad_edge_feat,
                    grad_node_feat,
                    grad_output,
                    ctx.graph.graph_csc,
                    ctx.dim_edge,
                    ctx.dim_node,
                    ctx.aggr,
                    out_pos,
                    stream_id=stream,
                )

            else:
                agg_concat_e2n_bwd(
                    grad_edge_feat,
                    grad_node_feat,
                    grad_output,
                    ctx.graph.graph_csc,
                    ctx.dim_edge,
                    ctx.dim_node,
                    ctx.aggr,
                    out_pos,
                    stream_id=stream,
                )
        else:
            grad_node_feat = maybe_zeros(
                **ctx.node_feat_meta_data,
                create_tensor=needs_grad_node,
            )
            grad_edge_feat = maybe_zeros(
                **ctx.edge_feat_meta_data,
                create_tensor=needs_grad_edge,
            )

        return grad_edge_feat, grad_node_feat, None, None


class _agg_concat_n2n_e2n_autograd(torch.autograd.Function):
    r"""Custom autograd function for simple aggregation using node features in
    an node-to-node reduction (n2n) and edge features in
    an edge-to-node reduction (e2n) on a graph while concatenating the
    original features of output at the end (agg_concat)."""

    @staticmethod
    def forward(ctx, node_feat, edge_feat, graph, aggr):
        dim_node = node_feat.size(-1)
        dim_edge = edge_feat.size(-1)
        agg_output = torch.empty(
            graph.num_dst_nodes,
            dim_node + dim_edge + dim_node,
            dtype=edge_feat.dtype,
            device=edge_feat.device,
        )

        out_pos = None
        if graph.num_dst_nodes > 0:
            if aggr in (AggOp.Max, AggOp.Min):
                idx_type = graph.graph_csc.__class__.__name__.split("_")[-1]
                out_pos = torch.empty(
                    graph.num_dst_nodes,
                    dim_node + dim_edge,
                    dtype=getattr(torch, idx_type),
                    device=edge_feat.device,
                )

            stream = torch.cuda.current_stream().cuda_stream

            node_feat = node_feat.detach().contiguous()
            edge_feat = edge_feat.detach().contiguous()
            agg_concat_n2n_e2n_fwd(
                agg_output,
                node_feat,
                edge_feat,
                graph.graph_csc,
                aggr,
                out_pos,
                stream_id=stream,
            )

        ctx.graph = graph
        ctx.dim_node = dim_node
        ctx.dim_edge = dim_edge
        ctx.aggr = aggr
        ctx.node_feat_meta_data = get_tensor_meta_data(node_feat)
        ctx.edge_feat_meta_data = get_tensor_meta_data(edge_feat)

        ctx.save_for_backward(out_pos)

        return agg_output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        grad_output = grad_output.detach().contiguous()
        (out_pos,) = ctx.saved_tensors
        needs_grad_node, needs_grad_edge, _, _ = ctx.needs_input_grad
        if not (needs_grad_node or needs_grad_edge):
            return None, None, None, None

        grad_output = grad_output.detach().contiguous()

        if ctx.graph.num_dst_nodes > 0:
            grad_node_feat = maybe_empty(
                **ctx.node_feat_meta_data, create_tensor=needs_grad_node
            )
            grad_edge_feat = maybe_empty(
                **ctx.edge_feat_meta_data, create_tensor=needs_grad_edge
            )

            stream = torch.cuda.current_stream().cuda_stream

            agg_concat_n2n_e2n_bwd(
                grad_node_feat,
                grad_edge_feat,
                grad_output,
                ctx.graph.graph_csc,
                ctx.dim_node,
                ctx.dim_edge,
                ctx.aggr,
                out_pos,
                stream_id=stream,
            )
        else:
            grad_node_feat = maybe_zeros(
                **ctx.node_feat_meta_data, create_tensor=needs_grad_node
            )
            grad_edge_feat = maybe_zeros(
                **ctx.edge_feat_meta_data, create_tensor=needs_grad_edge
            )

        return grad_node_feat, grad_edge_feat, None, None


class _agg_concat_n2n_e2n_bipartite_autograd(torch.autograd.Function):
    r"""Custom autograd function for simple aggregation using node features in
    an node-to-node reduction (n2n) and edge features in
    an edge-to-node reduction (e2n) on a graph while concatenating the
    original features of output at the end (agg_concat)."""

    @staticmethod
    def forward(ctx, src_feat, edge_feat, dst_feat, graph, aggr):
        dim_node = src_feat.size(-1)
        dim_edge = edge_feat.size(-1)
        dim_concat = dst_feat.size(-1)

        agg_output = torch.empty(
            graph.num_dst_nodes,
            dim_node + dim_edge + dim_concat,
            dtype=edge_feat.dtype,
            device=edge_feat.device,
        )

        out_pos = None
        if graph.num_dst_nodes > 0:
            if aggr in (AggOp.Max, AggOp.Min):
                idx_type = graph.graph_csc.__class__.__name__.split("_")[-1]
                out_pos = torch.empty(
                    graph.num_dst_nodes,
                    dim_node + dim_edge,
                    dtype=getattr(torch, idx_type),
                    device=edge_feat.device,
                )

            stream = torch.cuda.current_stream().cuda_stream

            src_feat = src_feat.detach().contiguous()
            edge_feat = edge_feat.detach().contiguous()
            dst_feat = dst_feat.detach().contiguous()

            agg_concat_n2n_e2n_bipartite_fwd(
                agg_output,
                src_feat,
                edge_feat,
                dst_feat,
                graph.graph_csc,
                aggr,
                out_pos,
                stream_id=stream,
            )

        ctx.graph = graph
        ctx.dim_node = dim_node
        ctx.dim_edge = dim_edge
        ctx.dim_concat = dim_concat
        ctx.aggr = aggr
        ctx.src_feat_meta_data = get_tensor_meta_data(src_feat)
        ctx.edge_feat_meta_data = get_tensor_meta_data(edge_feat)
        ctx.concat_feat_meta_data = get_tensor_meta_data(dst_feat)

        ctx.save_for_backward(out_pos)

        return agg_output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        grad_output = grad_output.detach().contiguous()
        (out_pos,) = ctx.saved_tensors
        needs_grad_node, needs_grad_edge, needs_grad_concat, _, _ = ctx.needs_input_grad
        if not (needs_grad_node or needs_grad_edge or needs_grad_concat):
            return None, None, None, None, None

        grad_output = grad_output.detach().contiguous()

        if ctx.graph.num_dst_nodes > 0:
            grad_src_feat = maybe_empty(
                **ctx.src_feat_meta_data, create_tensor=needs_grad_node
            )
            grad_edge_feat = maybe_empty(
                **ctx.edge_feat_meta_data, create_tensor=needs_grad_edge
            )
            grad_concat_feat = maybe_empty(
                **ctx.concat_feat_meta_data, create_tensor=needs_grad_concat
            )

            stream = torch.cuda.current_stream().cuda_stream

            agg_concat_n2n_e2n_bipartite_bwd(
                grad_src_feat,
                grad_edge_feat,
                grad_concat_feat,
                grad_output,
                ctx.graph.graph_csc,
                ctx.dim_node,
                ctx.dim_edge,
                ctx.dim_concat,
                ctx.aggr,
                out_pos,
                stream_id=stream,
            )
        else:
            grad_src_feat = maybe_zeros(
                **ctx.src_feat_meta_data, create_tensor=needs_grad_node
            )
            grad_edge_feat = maybe_zeros(
                **ctx.edge_feat_meta_data, create_tensor=needs_grad_edge
            )
            grad_concat_feat = maybe_zeros(
                **ctx.concat_feat_meta_data, create_tensor=needs_grad_concat
            )

        return grad_src_feat, grad_edge_feat, grad_concat_feat, None, None



# Copyright (c) 2023-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from typing import Optional
from torch.autograd.function import once_differentiable

from ..utils import (
    maybe_empty_like,
    maybe_detach,
    maybe_empty,
    get_tensor_meta_data,
    maybe_zeros,
    maybe_zeros_like,
)
from ..graph import CSC
from pylibcugraphops import mha_params
from pylibcugraphops.operators import (
    activation_params,
    mha_simple_n2n_fwd,
    mha_simple_n2n_bwd,
    mha_simple_n2n_efeat_fwd,
    mha_simple_n2n_efeat_bwd,
)


def mha_simple_n2n(
    key_emb: torch.Tensor,
    query_emb: torch.Tensor,
    value_emb: torch.Tensor,
    graph: CSC,
    num_heads: int = 1,
    concat_heads: bool = True,
    edge_emb: Optional[torch.Tensor] = None,
    norm_by_dim: bool = True,
    score_bias: Optional[torch.Tensor] = None,
    score_weight: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    r"""PyTorch autograd function for a multi-head attention layer (mha_simple)
    without using cudnn with an activation prior to the dot
    product but none afterwards in a node-to-node reduction (n2n).

    Parameters
    ----------
    key_emb : torch.Tensor
        The key embeddings of input nodes. Each row consists of concatenated features
        from different heads after the linear transformation.
        Shape: (n_src_nodes, dim_in) where dim_in = dim_head * num_heads, with
                dim_head being the feature dimension per head.

    query_emb : torch.Tensor
        The query embeddings of input nodes. Each row consists of concatenated features
        from different heads after the linear transformation.
        Shape: (n_dst_nodes, dim_in) where dim_in = dim_head * num_heads, with
                dim_head being the feature dimension per head.

    value_emb : torch.Tensor
        The value embeddings of input nodes. Each row consists of concatenated features
        from different heads after the linear transformation.
        Shape: (n_src_nodes, dim_in) where dim_in = dim_head * num_heads, with
                dim_head being the feature dimension per head.

    graph : CSC
        The graph used in for the operation.

    num_heads : int, default=1
        Number of heads in multi-head attention.

    concat_heads : bool, default=True
        Aggregated embeddings from each head are concatenated if `True` or
        averaged if `False`.

    edge_emb : Optional[torch.Tensor], default=None
        Optional input edge embeddings. Each row consists of concatenated
        features from different heads after the linear transformation.
        Shape: (n_edges, dim_in).

    norm_by_dim : bool, default=True
        flag indicating whether or not to normalize the result of
        the key-query dot-product with the square root of dim_in

    score_bias : Optional[torch.Tensor], default=None
        Optional additive bias which is added onto the result of the
        key-query dot-product before the softmax.

    score_weight : Optional[torch.Tensor], default=None
        Optional multplicative bias applied onto the result of the key-query
        dot-product before the softmax (prior to score_bias if also applied).

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape: (n_dst_nodes, dim_in) if `concat_heads` is `True`;
               (n_dst_nodes, dim_head) otherwise.
    """

    params = mha_params(activation_params(), num_heads, concat_heads)
    if not graph.is_bipartite:
        raise ValueError(
            "Expected bipartite with key_emb, query_emb,"
            " and value_emb being distinct buffers."
        )
    return _mha_simple_n2n_autograd.apply(
        key_emb,
        query_emb,
        value_emb,
        edge_emb,
        score_bias,
        score_weight,
        (graph, params, norm_by_dim),
    )


class _mha_simple_n2n_autograd(torch.autograd.Function):
    r"""Custom autograd function for a multi-head attention layer (mha_simple)
    without using cudnn (mha_gat_v2) operating on a graph
    in a node-to-node reduction (n2n)."""

    @staticmethod
    def forward(
        ctx, key_emb, query_emb, value_emb, edge_emb, score_bias, score_weight, params
    ):
        graph, params, norm_by_dim = params
        dim = value_emb.size(-1)
        dim_out = dim if params.concat_heads else (dim // params.num_heads)
        output = torch.empty(
            graph.num_dst_nodes, dim_out, dtype=value_emb.dtype, device=value_emb.device
        )
        sm_scores = torch.empty(
            2 if score_weight is None else 3,
            params.num_heads,
            graph.num_indices,
            dtype=value_emb.dtype,
            device=value_emb.device,
        )

        stream = torch.cuda.current_stream().cuda_stream

        key_emb = key_emb.detach().contiguous()
        query_emb = query_emb.detach().contiguous()
        value_emb = value_emb.detach().contiguous()
        edge_emb = maybe_detach(edge_emb)
        score_weight = maybe_detach(score_weight)
        score_bias = maybe_detach(score_bias)

        if graph.num_dst_nodes > 0:
            if edge_emb is None:
                mha_simple_n2n_fwd(
                    output,
                    sm_scores,
                    key_emb,
                    query_emb,
                    value_emb,
                    graph.graph_csc,
                    params,
                    norm_by_dim,
                    score_bias,
                    score_weight,
                    stream_id=stream,
                )
            else:
                mha_simple_n2n_efeat_fwd(
                    output,
                    sm_scores,
                    key_emb,
                    query_emb,
                    value_emb,
                    edge_emb,
                    graph.graph_csc,
                    params,
                    norm_by_dim,
                    score_bias,
                    score_weight,
                    stream_id=stream,
                )

        ctx.graph = graph
        ctx.params = params
        ctx.norm_by_dim = norm_by_dim
        ctx.score_bias_meta_data = get_tensor_meta_data(score_bias)
        ctx.save_for_backward(
            key_emb, query_emb, value_emb, edge_emb, sm_scores, score_weight
        )

        return output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        (
            key_emb,
            query_emb,
            value_emb,
            edge_emb,
            sm_scores,
            score_weight,
        ) = ctx.saved_tensors
        (
            needs_grad_key_emb,
            needs_grad_query_emb,
            needs_grad_value_emb,
            needs_grad_edge_emb,
            needs_grad_score_bias,
            needs_grad_score_weight,
            _,
        ) = ctx.needs_input_grad
        if not (
            needs_grad_key_emb
            or needs_grad_query_emb
            or needs_grad_value_emb
            or needs_grad_edge_emb
            or needs_grad_score_bias
            or needs_grad_score_weight
        ):
            return None, None, None, None, None, None

        grad_output = grad_output.detach().contiguous()

        if ctx.graph.num_dst_nodes > 0:
            grad_key_emb = maybe_empty_like(key_emb, needs_grad_key_emb)
            grad_query_emb = maybe_empty_like(query_emb, needs_grad_query_emb)
            grad_value_emb = maybe_empty_like(value_emb, needs_grad_value_emb)
            grad_sm_scores = torch.empty_like(sm_scores)

            # These two should be created if edge_emb and score_bias exist in
            # in the forward pass However, if they are passed as None, they also
            # don't require a gradient. Thus needs_grad_x should capture both
            # them existing in the forward pass but not requiring a gradient
            # and not existing in the first place.
            grad_edge_emb = maybe_empty_like(edge_emb, needs_grad_edge_emb)
            grad_score_bias = maybe_empty(
                **ctx.score_bias_meta_data, create_tensor=needs_grad_score_bias
            )
            grad_score_weight = maybe_empty_like(score_weight, needs_grad_score_weight)

            stream = torch.cuda.current_stream().cuda_stream

            if edge_emb is None:
                mha_simple_n2n_bwd(
                    grad_key_emb,
                    grad_query_emb,
                    grad_value_emb,
                    grad_sm_scores,
                    grad_output,
                    key_emb,
                    query_emb,
                    value_emb,
                    sm_scores,
                    ctx.graph.graph_csc,
                    ctx.params,
                    ctx.norm_by_dim,
                    score_weight,
                    grad_score_bias,
                    grad_score_weight,
                    stream_id=stream,
                )

            else:
                mha_simple_n2n_efeat_bwd(
                    grad_key_emb,
                    grad_query_emb,
                    grad_value_emb,
                    grad_edge_emb,
                    grad_sm_scores,
                    grad_output,
                    key_emb,
                    query_emb,
                    value_emb,
                    edge_emb,
                    sm_scores,
                    ctx.graph.graph_csc,
                    ctx.params,
                    ctx.norm_by_dim,
                    score_weight,
                    grad_score_bias,
                    grad_score_weight,
                    stream_id=stream,
                )
        else:
            grad_key_emb = maybe_zeros_like(key_emb, needs_grad_key_emb)
            grad_query_emb = maybe_zeros_like(query_emb, needs_grad_query_emb)
            grad_value_emb = maybe_zeros_like(value_emb, needs_grad_value_emb)
            grad_edge_emb = maybe_zeros_like(edge_emb, needs_grad_edge_emb)
            grad_score_bias = maybe_zeros(
                **ctx.score_bias_meta_data, create_tensor=needs_grad_score_bias
            )
            grad_score_weight = maybe_zeros_like(score_weight, needs_grad_score_weight)

        return (
            grad_key_emb,
            grad_query_emb,
            grad_value_emb,
            grad_edge_emb,
            grad_score_bias,
            grad_score_weight,
            None,
        )



# Copyright (c) 2023-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from typing import Optional, Tuple, Union
from torch.autograd.function import once_differentiable

from ..utils import (
    maybe_empty_like,
    maybe_empty,
    maybe_detach,
    get_tensor_meta_data,
    maybe_zeros_like,
    maybe_zeros,
)
from ..graph import CSC
from pylibcugraphops import ActivationOp, mha_params
from pylibcugraphops.operators import (
    activation_params,
    mha_gat_v2_n2n_fwd,
    mha_gat_v2_n2n_bwd,
    mha_gat_v2_n2n_efeat_fwd,
    mha_gat_v2_n2n_efeat_bwd,
)

_activation_ops = [x for x in dir(ActivationOp) if not x.startswith("__")]


def mha_gat_v2_n2n(
    feat: Union[torch.Tensor, Tuple[torch.Tensor]],
    attn_weights: torch.Tensor,
    graph: CSC,
    num_heads: int = 1,
    activation: str = "LeakyReLU",
    negative_slope: float = 0.2,
    concat_heads: bool = True,
    edge_feat: Optional[torch.Tensor] = None,
    deterministic_dgrad: Optional[bool] = False,
    deterministic_wgrad: Optional[bool] = False,
) -> torch.Tensor:
    """
    PyTorch autograd function for a multi-head attention layer (GAT-like)
    without using cudnn (mha_gat_v2) with an activation prior to the dot
    product but none afterwards in a node-to-node reduction (n2n).

    Parameters
    ----------
    feat : torch.Tensor | Tuple[torch.Tensor, torch.Tensor]
        The input node features.
        Depending on whether the graph is bipartite or not, this has
        to be a tuple of two tensors or a single tensor correspondingly.
        For bipartite graphs, the rows of each tensors consists of
        concatenated features from different heads after the linear transformation.
        Shapes: ``(n_src_nodes, dim_in), (n_dst_nodes, dim_in)``,
        where ``dim_in = dim_head * num_heads``, with ``dim_head`` being the
        feature dimension per head.
        For non-bipartite graphs, the rows of the single tensor consists of
        concatenated features from different heads after the linear transformation.
        Shape: ``(n_src_nodes, dim_in)``, where ``dim_in = dim_head * num_heads``,
        with ``dim_head`` being the feature dimension per head.

    attn_weights : torch.Tensor
        The attention weight.
        Shape: ``(dim_in,)``.

    graph : CSC
        The graph used in for the operation.

    num_heads : int, default=1
        Number of heads in multi-head attention.

    activation : str, default="LeakyReLU"
        The activation function used in the attention mechanism. Choose from
        ``"ELU"``, ``"LeakyReLU"``, ``"Linear"``, ``"ReLU"``, ``"Scalar"``,
        ``"Sigmoid"`` and ``"Tanh"``.

    negative_slope : float, default=0.2
        LeakyReLU angle of negative slope. Only effective when ``activation`` is
        ``"LeakyReLU"``.

    concat_heads : bool, default=True
        Aggregated embeddings from each head are concatenated if ``True`` or
        averaged if ``False``.

    edge_feat : Optional[torch.Tensor], default=None
        Optional input edge features. Each row consists of concatenated features
        from different heads after the linear transformation.
        Shape: ``(n_edges, dim_in)``.

    deterministic_dgrad : Optional[bool], default=False
        Optional flag indicating whether the gradient w.r.t. the features
        is computed deterministically using a dedicated workspace buffer.

    deterministic_wgrad : Optional[bool], default=False
        Optional flag indicating whether the gradient w.r.t. the weights
        is computed deterministically using a dedicated workspace buffer.

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape for ``concat_heads=True``: ``(n_dst_nodes, dim_in)``;
        Shape for ``concat_heads=False``: ``(n_dst_nodes, dim_head)``.
    """
    if activation not in _activation_ops:
        raise ValueError(
            f"Supported activation functions: {_activation_ops}, "
            f"but got '{activation}'."
        )
    act_params = activation_params(negative_slope, getattr(ActivationOp, activation))
    params = mha_params(act_params, num_heads, concat_heads)
    if graph.is_bipartite:
        assert isinstance(feat, Tuple) and len(feat) == 2, (
            f"Expected tuple of source and destination node features,"
            f" got {type(feat)} instead."
        )
        return _mha_gat_v2_n2n_bipartite_autograd.apply(
            feat[0],
            feat[1],
            edge_feat,
            attn_weights,
            graph,
            params,
            deterministic_dgrad,
            deterministic_wgrad,
        )

    assert isinstance(
        feat, torch.Tensor
    ), f"Expected torch.Tensor of node features, got {type(feat)} instead."
    return _mha_gat_v2_n2n_autograd.apply(
        feat,
        edge_feat,
        attn_weights,
        graph,
        params,
        deterministic_dgrad,
        deterministic_wgrad,
    )


class _mha_gat_v2_n2n_autograd(torch.autograd.Function):
    """
    Custom autograd function for a multi-head attention layer (GAT-like)
    without using cudnn (mha_gat_v2) with an activation prior to the dot
    product but none afterwards operating on a graph in a node-to-node
    reduction (n2n).
    """

    @staticmethod
    def forward(
        ctx,
        feat,
        edge_feat,
        attn_weights,
        graph,
        params,
        deterministic_dgrad,
        deterministic_wgrad,
    ):
        dim = feat.size(-1)
        feat = feat.detach().contiguous()
        attn_weights = attn_weights.detach().contiguous()
        edge_feat = maybe_detach(edge_feat)
        dim_out = dim if params.concat_heads else (dim // params.num_heads)
        output = torch.empty(
            graph.num_dst_nodes, dim_out, dtype=feat.dtype, device=feat.device
        )
        sm_scores = torch.empty(
            2,
            params.num_heads,
            graph.num_indices,
            dtype=feat.dtype,
            device=feat.device,
        )
        act_scores = torch.empty(
            graph.num_indices, dim, dtype=feat.dtype, device=feat.device
        )

        stream = torch.cuda.current_stream().cuda_stream

        if graph.num_dst_nodes > 0:
            if edge_feat is None:
                mha_gat_v2_n2n_fwd(
                    output,
                    sm_scores,
                    act_scores,
                    feat,
                    attn_weights,
                    graph.graph_csc,
                    params,
                    stream_id=stream,
                )
            else:
                mha_gat_v2_n2n_efeat_fwd(
                    output,
                    sm_scores,
                    act_scores,
                    feat,
                    edge_feat,
                    attn_weights,
                    graph.graph_csc,
                    params,
                    stream_id=stream,
                )

        ctx.graph = graph
        ctx.params = params
        ctx.deterministic_dgrad = deterministic_dgrad
        ctx.deterministic_wgrad = deterministic_wgrad
        ctx.save_for_backward(feat, edge_feat, attn_weights, sm_scores, act_scores)

        return output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        feat, edge_feat, attn_weights, sm_scores, act_scores = ctx.saved_tensors
        (
            needs_grad_feat,
            needs_grad_efeat,
            needs_grad_weights,
            _,
            _,
            _,
            _,
        ) = ctx.needs_input_grad
        if not (needs_grad_feat or needs_grad_weights or needs_grad_efeat):
            return None, None, None, None, None, None, None

        grad_output = grad_output.detach().contiguous()

        if ctx.graph.num_dst_nodes > 0:
            grad_feat = maybe_empty_like(feat, needs_grad_feat)
            # needs_grad_efeat is also False if edge_feat is None
            # thus capturing all possible cases
            grad_edge_feat = maybe_empty_like(edge_feat, needs_grad_efeat)
            grad_attn_weights = maybe_empty_like(attn_weights, needs_grad_weights)
            grad_wspace_feat = maybe_empty(
                (ctx.graph.num_indices, feat.size(1)),
                create_tensor=(needs_grad_feat and ctx.deterministic_dgrad),
                dtype=feat.dtype,
                device=feat.device,
            )
            grad_wspace_weight = maybe_empty(
                (ctx.graph.num_dst_nodes, attn_weights.size(0)),
                create_tensor=(needs_grad_weights and ctx.deterministic_wgrad),
                dtype=attn_weights.dtype,
                device=attn_weights.device,
            )
            # grad_sm_scores is basically a workspace, thus always required
            grad_sm_scores = torch.empty_like(sm_scores)

            stream = torch.cuda.current_stream().cuda_stream

            if edge_feat is None:
                mha_gat_v2_n2n_bwd(
                    grad_feat,
                    grad_attn_weights,
                    grad_sm_scores,
                    grad_output,
                    feat,
                    attn_weights,
                    sm_scores,
                    act_scores,
                    ctx.graph.graph_csc,
                    ctx.params,
                    grad_wspace_feat,
                    grad_wspace_weight,
                    stream_id=stream,
                )

            else:
                mha_gat_v2_n2n_efeat_bwd(
                    grad_feat,
                    grad_edge_feat,
                    grad_attn_weights,
                    grad_sm_scores,
                    grad_output,
                    feat,
                    attn_weights,
                    sm_scores,
                    act_scores,
                    ctx.graph.graph_csc,
                    ctx.params,
                    grad_wspace_feat,
                    grad_wspace_weight,
                    stream_id=stream,
                )
        else:
            grad_feat = maybe_zeros_like(feat, needs_grad_feat)
            grad_edge_feat = maybe_zeros_like(edge_feat, needs_grad_efeat)
            grad_attn_weights = maybe_zeros_like(attn_weights, needs_grad_weights)

        return grad_feat, grad_edge_feat, grad_attn_weights, None, None, None, None


class _mha_gat_v2_n2n_bipartite_autograd(torch.autograd.Function):
    """
    Custom autograd function for a multi-head attention layer (GAT-like)
    without using cudnn (mha_gat_v2) with an activation prior to the dot
    product but none afterwards operating on bipartite graphs in a node-to-node
    reduction (n2n).
    """

    @staticmethod
    def forward(
        ctx,
        src_feat,
        dst_feat,
        edge_feat,
        attn_weights,
        graph,
        params,
        deterministic_dgrad,
        deterministic_wgrad,
    ):
        dim = src_feat.size(-1)
        src_feat = src_feat.detach().contiguous()
        dst_feat = dst_feat.detach().contiguous()
        attn_weights = attn_weights.detach().contiguous()
        edge_feat = maybe_detach(edge_feat)
        dim_out = dim if params.concat_heads else (dim // params.num_heads)
        output = torch.empty(
            graph.num_dst_nodes, dim_out, dtype=src_feat.dtype, device=src_feat.device
        )
        sm_scores = torch.empty(
            2,
            params.num_heads,
            graph.num_indices,
            dtype=src_feat.dtype,
            device=src_feat.device,
        )
        act_scores = torch.empty(
            graph.num_indices, dim, dtype=src_feat.dtype, device=src_feat.device
        )

        stream = torch.cuda.current_stream().cuda_stream

        if graph.num_dst_nodes > 0:
            if edge_feat is None:
                mha_gat_v2_n2n_fwd(
                    output,
                    sm_scores,
                    act_scores,
                    src_feat,
                    dst_feat,
                    attn_weights,
                    graph.graph_csc,
                    params,
                    stream_id=stream,
                )
            else:
                mha_gat_v2_n2n_efeat_fwd(
                    output,
                    sm_scores,
                    act_scores,
                    src_feat,
                    dst_feat,
                    edge_feat,
                    attn_weights,
                    graph.graph_csc,
                    params,
                    stream_id=stream,
                )

        ctx.graph = graph
        ctx.params = params
        ctx.deterministic_dgrad = deterministic_dgrad
        ctx.deterministic_wgrad = deterministic_wgrad
        ctx.dst_feat_meta_data = get_tensor_meta_data(dst_feat)
        ctx.save_for_backward(src_feat, edge_feat, attn_weights, sm_scores, act_scores)

        return output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        src_feat, edge_feat, attn_weights, sm_scores, act_scores = ctx.saved_tensors
        (
            needs_grad_src_feat,
            needs_grad_dst_feat,
            needs_grad_efeat,
            needs_grad_weights,
            _,
            _,
            _,
            _,
        ) = ctx.needs_input_grad
        if not (
            needs_grad_src_feat
            or needs_grad_dst_feat
            or needs_grad_weights
            or needs_grad_efeat
        ):
            return None, None, None, None, None, None, None, None

        grad_output = grad_output.detach().contiguous()

        if ctx.graph.num_dst_nodes > 0:
            grad_src_feat = maybe_empty_like(src_feat, needs_grad_src_feat)
            grad_dst_feat = maybe_empty(
                **ctx.dst_feat_meta_data, create_tensor=needs_grad_dst_feat
            )
            # needs_grad_efeat is also False if edge_feat is None
            # thus capturing all possible cases
            grad_edge_feat = maybe_empty_like(edge_feat, needs_grad_efeat)
            grad_attn_weights = maybe_empty_like(attn_weights, needs_grad_weights)
            grad_wspace_feat = maybe_empty(
                (ctx.graph.num_indices, src_feat.size(1)),
                create_tensor=(needs_grad_src_feat and ctx.deterministic_dgrad),
                dtype=src_feat.dtype,
                device=src_feat.device,
            )
            grad_wspace_weights = maybe_empty(
                (ctx.graph.num_dst_nodes, attn_weights.size(0)),
                create_tensor=(needs_grad_weights and ctx.deterministic_wgrad),
                dtype=attn_weights.dtype,
                device=attn_weights.device,
            )
            # grad_sm_scores is basically a workspace, thus always required
            grad_sm_scores = torch.empty_like(sm_scores)

            stream = torch.cuda.current_stream().cuda_stream

            if edge_feat is None:
                mha_gat_v2_n2n_bwd(
                    grad_src_feat,
                    grad_dst_feat,
                    grad_attn_weights,
                    grad_sm_scores,
                    grad_output,
                    src_feat,
                    attn_weights,
                    sm_scores,
                    act_scores,
                    ctx.graph.graph_csc,
                    ctx.params,
                    grad_wspace_feat,
                    grad_wspace_weights,
                    stream_id=stream,
                )

            else:
                mha_gat_v2_n2n_efeat_bwd(
                    grad_src_feat,
                    grad_dst_feat,
                    grad_edge_feat,
                    grad_attn_weights,
                    grad_sm_scores,
                    grad_output,
                    src_feat,
                    attn_weights,
                    sm_scores,
                    act_scores,
                    ctx.graph.graph_csc,
                    ctx.params,
                    grad_wspace_feat,
                    grad_wspace_weights,
                    stream_id=stream,
                )
        else:
            grad_src_feat = maybe_zeros_like(src_feat, needs_grad_src_feat)
            grad_dst_feat = maybe_zeros(
                **ctx.dst_feat_meta_data, create_tensor=needs_grad_dst_feat
            )
            grad_edge_feat = maybe_zeros_like(edge_feat, needs_grad_efeat)
            grad_attn_weights = maybe_zeros_like(attn_weights, needs_grad_weights)

        return (
            grad_src_feat,
            grad_dst_feat,
            grad_edge_feat,
            grad_attn_weights,
            None,
            None,
            None,
            None,
        )



# Copyright (c) 2022-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import warnings
from typing import Tuple, Any

import torch
from torch import Tensor
import pylibcugraphops


class CSC:
    def __init__(
        self,
        offsets: Tensor,
        indices: Tensor,
        num_src_nodes: int,
        map_csc_to_coo: Tensor | None = None,
        rev_offsets: Tensor | None = None,
        rev_indices: Tensor | None = None,
        map_rev_to_coo: Tensor | None = None,
        map_dst_to_src: Tensor | None = None,
        dst_max_in_degree: int = -1,
        reverse_graph_bwd: bool = False,
        is_bipartite: bool = False,
    ):
        self.map_dst_to_src = map_dst_to_src
        self.num_src_nodes = num_src_nodes
        self.num_dst_nodes = offsets.size(0) - 1
        self.num_indices = indices.size(0)

        self.offsets = offsets
        self.indices = indices
        self.map_csc_to_coo = map_csc_to_coo

        self._rev_cache_filled = False
        if reverse_graph_bwd:
            if (
                isinstance(rev_offsets, Tensor)
                and isinstance(rev_indices, Tensor)
                and isinstance(map_rev_to_coo, Tensor)
            ):
                self.rev_offsets = rev_offsets
                self.rev_indices = rev_indices
                self.map_rev_to_coo = map_rev_to_coo
                self._rev_cache_filled = True
            else:
                self._create_rev_buffers()
        else:
            self.rev_offsets = self.rev_indices = self.map_rev_to_coo = None
        self.reverse_graph_bwd = reverse_graph_bwd

        self.is_bipartite = is_bipartite

        if map_dst_to_src is not None:
            raise NotImplementedError("map_dst_to_src is currently not supported")

        # TODO remove warning once we again have support for these
        if dst_max_in_degree > 0:
            warnings.warn("dst_max_in_degree currently has no effect")
        self.dst_max_in_degree = dst_max_in_degree

        self._build_graph()

        if self.reverse_graph_bwd:
            self.add_reverse_graph()

    def _create_rev_buffers(self):
        self.rev_offsets = torch.empty(
            self.num_src_nodes + 1,
            device=self.offsets.device,
            dtype=self.offsets.dtype,
        )
        self.rev_indices = torch.empty_like(self.indices)
        self.map_rev_to_coo = torch.empty_like(self.indices)

    def _build_graph(self):
        if self.is_bipartite:
            make_graph_func = pylibcugraphops.make_bipartite_csc
        else:
            make_graph_func = pylibcugraphops.make_csc

        if not self.reverse_graph_bwd:
            self.graph_csc = make_graph_func(
                self.offsets,
                self.indices,
                self.num_src_nodes,
                self.map_csc_to_coo,
            )
        else:
            self.graph_csc = make_graph_func(
                self.offsets,
                self.indices,
                self.num_src_nodes,
                self.map_csc_to_coo,
                self.rev_offsets,
                self.rev_indices,
                self.map_rev_to_coo,
            )

    def add_reverse_graph(self):
        if self._rev_cache_filled:
            return

        # skip reversing empty graphs to avoid zero-grid kernel launch
        if self.indices.numel() == 0:
            return

        if self.reverse_graph_bwd is False:
            self.reverse_graph_bwd = True
            self._create_rev_buffers()
            self._build_graph()

        stream = torch.cuda.current_stream().cuda_stream
        # call reverse to get necessary workspace size and
        # create necessary workspace buffers
        node_counts = torch.empty(
            self.num_src_nodes,
            dtype=self.offsets.dtype,
            device=self.offsets.device,
        )
        cub_workspace = None
        cub_workspace_size = pylibcugraphops.reverse_graph(
            self.graph_csc, node_counts, cub_workspace, 0, stream
        )
        cub_workspace = torch.empty(
            cub_workspace_size, dtype=torch.int8, device=self.offsets.device
        )
        # actually reverse graph (in-place)
        pylibcugraphops.reverse_graph(
            self.graph_csc, node_counts, cub_workspace, cub_workspace_size, stream
        )

        self._rev_cache_filled = True

    def to(self, *args, **kwargs):
        self.offsets = self.offsets.to(*args, **kwargs)
        self.indices = self.indices.to(*args, **kwargs)
        if self.map_dst_to_src is not None:
            self.map_dst_to_src = self.map_dst_to_src.to(*args, **kwargs)
        if self.map_csc_to_coo is not None:
            self.map_csc_to_coo = self.map_csc_to_coo.to(*args, **kwargs)
        if self.map_rev_to_coo is not None:
            self.map_rev_to_coo = self.map_rev_to_coo.to(*args, **kwargs)
        if self.reverse_graph_bwd:
            self.rev_offsets = self.rev_offsets.to(*args, **kwargs)
            self.rev_indices = self.rev_indices.to(*args, **kwargs)

        self._build_graph()

    def __to__(self, *args, **kwargs):
        self.to(*args, **kwargs)

    @classmethod
    def from_edge_index(
        cls,
        edge_index: Any,
        reverse_graph_bwd: bool = False,
        is_bipartite: bool = False,
    ):
        """Create CSC graph from a torch_geometric.EdgeIndex object."""
        num_src_nodes = edge_index.sparse_size(0)
        edge_index = edge_index.sort_by("col")[0]
        colptr, row = edge_index.get_csc()[0]

        if not reverse_graph_bwd:
            return cls(
                colptr,
                row,
                num_src_nodes,
                reverse_graph_bwd=False,
                is_bipartite=is_bipartite,
            )

        (rowptr, col), perm = edge_index.get_csr()
        perm = perm.to(dtype=edge_index.dtype)
        return cls(
            colptr,
            row,
            num_src_nodes,
            rev_offsets=rowptr,
            rev_indices=col,
            map_rev_to_coo=perm,
            reverse_graph_bwd=True,
            is_bipartite=is_bipartite,
        )


class HeteroCSC(CSC):
    def __init__(
        self,
        offsets: Tensor,
        indices: Tensor,
        edge_types: Tensor,
        num_src_nodes: int,
        num_edge_types: int,
        num_node_types: int = 0,
        node_types: Tensor | Tuple[Tensor, Tensor] | None = None,
        map_csc_to_coo: Tensor | None = None,
        rev_offsets: Tensor | None = None,
        rev_indices: Tensor | None = None,
        map_rev_to_coo: Tensor | None = None,
        map_dst_to_src: Tensor | None = None,
        dst_max_in_degree: int = -1,
        reverse_graph_bwd: bool = False,
        is_bipartite: bool = False,
    ):
        if is_bipartite:
            if num_node_types > 0:
                assert isinstance(node_types, Tuple), (
                    "Expected tuple of node types for a bipartite graph"
                    f", but got {type(node_types)} instead."
                )

        else:
            if num_node_types > 0:
                assert isinstance(node_types, Tensor), (
                    "Expected node_types to be of type Tensor"
                    f", but got {type(node_types)} instead."
                )

        self.edge_types = edge_types.to(torch.int32)
        self.num_node_types = num_node_types
        self.num_edge_types = num_edge_types
        if node_types is None:
            self.node_types = None
        elif isinstance(node_types, Tensor):
            self.node_types = node_types.to(torch.int32)
        else:
            self.node_types = (
                node_types[0].to(torch.int32),
                node_types[1].to(torch.int32),
            )

        super().__init__(
            offsets,
            indices,
            num_src_nodes,
            map_csc_to_coo=map_csc_to_coo,
            rev_offsets=rev_offsets,
            rev_indices=rev_indices,
            map_rev_to_coo=map_rev_to_coo,
            map_dst_to_src=map_dst_to_src,
            dst_max_in_degree=dst_max_in_degree,
            reverse_graph_bwd=reverse_graph_bwd,
            is_bipartite=is_bipartite,
        )

    def _build_graph(self):
        if self.is_bipartite:
            src_node_types, dst_node_types = self.node_types
            self.graph_csc = pylibcugraphops.make_bipartite_csc_hg(
                self.offsets,
                self.indices,
                self.num_src_nodes,
                self.num_node_types,
                self.num_edge_types,
                src_node_types,
                dst_node_types,
                self.edge_types,
                self.map_csc_to_coo,
                self.rev_offsets,
                self.rev_indices,
                self.map_rev_to_coo,
            )

        else:
            self.graph_csc = pylibcugraphops.make_csc_hg(
                self.offsets,
                self.indices,
                self.num_src_nodes,
                self.num_node_types,
                self.num_edge_types,
                self.node_types,
                self.edge_types,
                self.map_csc_to_coo,
                self.rev_offsets,
                self.rev_indices,
                self.map_rev_to_coo,
            )

    def to(self, *args, **kwargs):
        self.edge_types = self.edge_types.to(*args, **kwargs).to(torch.int32)
        if self.node_types is not None:
            if isinstance(self.node_types, Tensor):
                self.node_types = self.node_types.to(*args, **kwargs).to(torch.int32)
            else:
                self.node_types = (
                    self.node_types[0].to(*args, **kwargs).to(torch.int32),
                    self.node_types[1].to(*args, **kwargs).to(torch.int32),
                )
        super().to(*args, **kwargs)



# Copyright (c) 2023-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import pylibcugraphops.operators as ops


class transpose_segment_to_contiguous_autograd(torch.autograd.Function):
    @staticmethod
    def forward(ctx, tensor: torch.Tensor, segment_info: torch.Tensor):
        segment_info = segment_info.detach().contiguous()
        ctx.save_for_backward(segment_info)
        tensor_transpose = torch.empty_like(tensor)
        stream = torch.cuda.current_stream().cuda_stream

        if tensor.numel() > 0:
            ops.segmented_transpose(
                tensor_transpose.detach().contiguous(),
                tensor.detach().contiguous(),
                segment_info,
                True,
                stream,
            )
        return tensor_transpose

    @staticmethod
    def backward(ctx, grad_out: torch.Tensor):
        (segment_info,) = ctx.saved_tensors
        grad_tensor = transpose_segment_to_strided_autograd.apply(
            grad_out, segment_info
        )
        return grad_tensor, None


class transpose_segment_to_strided_autograd(torch.autograd.Function):
    @staticmethod
    def forward(ctx, tensor: torch.Tensor, segment_info: torch.Tensor):
        segment_info = segment_info.detach().contiguous()
        ctx.save_for_backward(segment_info)
        tensor_transpose = torch.empty_like(tensor)
        stream = torch.cuda.current_stream().cuda_stream

        if tensor.numel() > 0:
            ops.segmented_transpose(
                tensor_transpose.detach().contiguous(),
                tensor.detach().contiguous(),
                segment_info,
                False,
                stream,
            )
        return tensor_transpose

    @staticmethod
    def backward(ctx, grad_out: torch.Tensor):
        (segment_info,) = ctx.saved_tensors
        grad_tensor = transpose_segment_to_contiguous_autograd.apply(
            grad_out, segment_info
        )
        return grad_tensor, None


def transpose_segment_to_contiguous(
    tensor: torch.Tensor, segment_info: torch.Tensor
) -> torch.Tensor:
    if segment_info is None:
        return tensor
    return transpose_segment_to_contiguous_autograd.apply(tensor, segment_info)


def transpose_segment_to_strided(
    tensor: torch.Tensor, segment_info: torch.Tensor
) -> torch.Tensor:
    if segment_info is None:
        return tensor
    return transpose_segment_to_strided_autograd.apply(tensor, segment_info)



# Copyright (c) 2022-2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
from torch.autograd.function import once_differentiable

from ..utils import maybe_empty_like, maybe_detach
from ..graph import HeteroCSC
from pylibcugraphops.operators import (
    agg_hg_basis_n2n_post_fwd,
    agg_hg_basis_n2n_post_bwd,
)


def agg_hg_basis_n2n_post(
    feat: torch.Tensor,
    weights_comb: Optional[torch.Tensor],
    graph: HeteroCSC,
    concat_own: bool = False,
    norm_by_out_degree: bool = False,
) -> torch.Tensor:
    """
    PyTorch autograd function for node-to-node RGCN-like basis regularized
    aggregation, with features being transformed after (post) this aggregation.

    Parameters
    ----------
    feat : torch.Tensor
        The input node features.
        Shape: ``(n_src_nodes, dim_in)``.

    weights_comb : Optional[torch.Tensor], default=None
        The combination weights. If `None`, no weights are used and features of
        different edge types are aggregated into separate output dimensions.
        Shape: ``(n_edge_types, n_bases)`` if provided.

    graph : HeteroCSC
        The graph used for the operation.

    concat_own : bool, default=False
        Concatenate output node embeddings in the aggregation.

    norm_by_out_degree : bool, default=False
        If set, output embeddings are normed by the degree of the output node
        per edge type.

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape for ``concat_own=False``: ``(n_dst_nodes, dim_out)``.
        Shape for ``concat_own=True``: ``(n_dst_nodes, dim_out + dim_in)``,
        where ``dim_out = dim_in * n_bases`` if ``weights_comb`` is provided,
        and ``dim_out = dim_in * n_edge_types`` if not provided.
    """
    options = (concat_own, norm_by_out_degree)

    if graph.is_bipartite:
        raise NotImplementedError(
            "This operator currently does not implement support for bipartite graphs."
        )

    return _agg_hg_basis_n2n_post_autograd.apply(feat, weights_comb, graph, options)


class _agg_hg_basis_n2n_post_autograd(torch.autograd.Function):
    """
    Custom autograd function for node-to-node full-graph RGCN-like basis
    regularized aggregation, with features being transformed after (post)
    this aggregation.
    """

    @staticmethod
    def forward(ctx, feat, weights_comb, graph, options):
        concat_own, norm_by_out_degree = options
        dim = feat.size(-1)
        n_bases = (
            graph.num_edge_types if weights_comb is None else weights_comb.size(-1)
        )
        dim_out = dim * (n_bases + (1 if concat_own else 0))

        agg_output = torch.empty(
            graph.num_dst_nodes, dim_out, dtype=feat.dtype, device=feat.device
        )

        stream = torch.cuda.current_stream().cuda_stream

        feat = feat.detach().contiguous()
        weights_comb = maybe_detach(weights_comb)
        agg_hg_basis_n2n_post_fwd(
            agg_output,
            feat,
            weights_comb,
            graph.graph_csc,
            concat_own,
            norm_by_out_degree,
            stream_id=stream,
        )

        ctx.graph = graph
        ctx.concat_own = concat_own
        ctx.norm_by_out_degree = norm_by_out_degree
        ctx.save_for_backward(feat, weights_comb)

        return agg_output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        grad_output = grad_output.detach().contiguous()
        feat, weights_comb = ctx.saved_tensors
        needs_grad_feat, needs_grad_w, _, _ = ctx.needs_input_grad
        if not (needs_grad_feat or needs_grad_w):
            return None, None, None, None

        grad_feat = maybe_empty_like(feat, needs_grad_feat)

        stream = torch.cuda.current_stream().cuda_stream
        if weights_comb is None:
            grad_weights_comb = None
        else:
            grad_weights_comb = maybe_empty_like(weights_comb, needs_grad_w)

        agg_hg_basis_n2n_post_bwd(
            grad_feat,
            grad_weights_comb,
            grad_output,
            feat,
            weights_comb,
            ctx.graph.graph_csc,
            ctx.concat_own,
            ctx.norm_by_out_degree,
            stream_id=stream,
        )

        return grad_feat, grad_weights_comb, None, None



# Copyright (c) 2023-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn
from functools import cmp_to_key

import pylibcugraphops

from pylibcugraphops.pytorch.utils import (
    maybe_detach,
    get_operator_from_module,
)

import pylibcugraphops.operators as ops


def _cmp_indices_tuples(a, b):
    if a[1][0] > b[1][0]:
        return 1
    elif a[1][0] == b[1][0]:
        if len(a[1]) < len(b[1]):
            return -1
        elif len(a[1]) == len(b[1]):
            return 0
        else:
            return 1
    else:
        return -1


def _matricization(path_vals, path_indices, num_out_segments, correlation):
    path_padded_vals = torch.tensor([], dtype=torch.float64)
    path_padded_indices = torch.tensor([], dtype=torch.int32)

    path_offsets = torch.zeros(2 * num_out_segments, dtype=torch.int32)

    sorted_vals_indices = sorted(
        zip(path_vals, path_indices), key=cmp_to_key(_cmp_indices_tuples)
    )

    curr_output = -1
    path_start, path_end, path_num = 0, 0, 0

    for path_val, path_index in sorted_vals_indices:

        if len(path_index) > correlation + 2:
            raise AssertionError("Invalid length of index tuple")

        if curr_output == -1:

            curr_output = path_index[0]
            path_num = 1

        elif curr_output == path_index[0]:

            path_num = path_num + 1

        elif curr_output != path_index[0]:

            path_end = path_start + path_num
            path_offsets[2 * curr_output + 0] = path_start
            path_offsets[2 * curr_output + 1] = path_end

            num_padded = ((path_num + 3) // 4) * 4 - path_num

            for i in range(0, num_padded):
                path_padded_vals = torch.cat(
                    (path_padded_vals, torch.tensor([0], dtype=torch.float64))
                )
                path_padded_indices = torch.cat(
                    (
                        path_padded_indices,
                        torch.tensor([0] * (correlation + 1), dtype=torch.int32),
                    )
                )

            path_num = 1
            path_start = path_end + num_padded
            path_end = path_start

            curr_output = path_index[0]

        else:
            raise AssertionError(
                "Unexpected value of curr_output and path index",
                curr_output,
                path_index,
            )

        path_padded_vals = torch.cat(
            (path_padded_vals, torch.tensor([path_val], dtype=torch.float64))
        )

        index = (
            path_index[1:-1]
            + [-1] * (correlation - len(path_index[1:-1]))
            + [path_index[-1]]
        )

        path_padded_indices = torch.cat(
            (path_padded_indices, torch.tensor(index, dtype=torch.int32))
        )

    if curr_output != -1:
        num_padded = ((path_num + 3) // 4) * 4 - path_num

        for i in range(0, num_padded):
            path_padded_vals = torch.cat(
                (path_padded_vals, torch.tensor([0], dtype=torch.float64))
            )
            path_padded_indices = torch.cat(
                (
                    path_padded_indices,
                    torch.tensor([0] * (correlation + 1), dtype=torch.int32),
                )
            )

        path_end = path_start + path_num
        path_offsets[2 * curr_output + 0] = path_start
        path_offsets[2 * curr_output + 1] = path_end

    return path_padded_vals, path_padded_indices, path_offsets


class SymmetricContractionFwdInfo(nn.Module):
    def __init__(self, path_vals, path_indices, num_out_segments, correlation, dtype):
        super().__init__()

        self.num_out_segments = num_out_segments

        path_vals, path_indices, path_offsets = _matricization(
            path_vals, path_indices, num_out_segments, correlation
        )

        path_vals = path_vals.to(dtype).contiguous()

        self.register_buffer("path_vals", path_vals)

        path_indices = path_indices.to(torch.int16).contiguous()
        self.register_buffer("path_indices", path_indices)

        path_offsets = path_offsets.to(torch.int32).flatten().contiguous()
        self.register_buffer("path_offsets", path_offsets)

        self.make_clebsch_gordan_tensor = get_operator_from_module(
            pylibcugraphops,
            "make_clebsch_gordan_tensor",
            (dtype,),
        )

    def forward(self):
        raise NotImplementedError(
            "SymmetricContractionFwdInfo module is not intended to be called."
        )

    def c_type(self):
        return self.make_clebsch_gordan_tensor(
            self.path_vals, self.path_indices, self.path_offsets, self.num_out_segments
        )


def _diff_index_tuple_list_tensor(path_vals, path_indices, num_in_segments):
    new_path_vals, new_path_indices = [], []

    for path_val, path_index in zip(path_vals, path_indices):

        for i_segment in range(0, num_in_segments):

            tmp = path_index[1:-1]
            num = tmp.count(i_segment)

            if num == 0:
                continue

            tmp.pop(tmp.index(i_segment))
            new_path_index = [i_segment] + [path_index[0]] + tmp + [path_index[-1]]
            new_path_val = num * path_val

            new_path_vals.append(new_path_val)
            new_path_indices.append(new_path_index)

    return new_path_vals, new_path_indices


def _diff_index_tuple_list_weight(path_vals, path_indices):

    new_path_indices = []
    for path_index in path_indices:
        index = [path_index[-1]] + path_index[1:-1] + [path_index[0]]
        new_path_indices.append(index)

    return path_vals, new_path_indices


class SymmetricContractionBwdInfo(nn.Module):
    def __init__(
        self,
        path_vals,
        path_indices,
        num_in_segments,
        num_coupling_paths,
        correlation,
        dtype=torch.float32,
    ):
        super().__init__()

        self.num_in_segments = num_in_segments
        self.num_coupling_paths = num_coupling_paths

        # for weight derivatives
        new_path_vals, new_path_indices = _diff_index_tuple_list_weight(
            path_vals, path_indices
        )

        path_padded_vals, path_padded_indices, path_offsets = _matricization(
            new_path_vals, new_path_indices, num_coupling_paths, correlation
        )

        path_vals_w = path_padded_vals.to(dtype).contiguous()
        self.register_buffer("path_vals_w", path_vals_w)

        path_indices_w = path_padded_indices.to(torch.int16).contiguous()
        self.register_buffer("path_indices_w", path_indices_w)

        path_offsets_w = path_offsets.to(torch.int32).contiguous()
        self.register_buffer("path_offsets_w", path_offsets_w)

        # for input tensor derivatives
        new_path_vals, new_path_indices = _diff_index_tuple_list_tensor(
            path_vals, path_indices, num_in_segments
        )

        path_padded_vals, path_padded_indices, path_offsets = _matricization(
            new_path_vals, new_path_indices, num_in_segments, correlation
        )

        path_vals = path_padded_vals.to(dtype).contiguous()
        self.register_buffer("path_vals", path_vals)

        path_indices = path_padded_indices.to(torch.int16).contiguous()
        self.register_buffer("path_indices", path_indices)

        path_offsets = path_offsets.to(torch.int32).flatten().contiguous()
        self.register_buffer("path_offsets", path_offsets)

        self.make_clebsch_gordan_tensor = get_operator_from_module(
            pylibcugraphops,
            "make_clebsch_gordan_tensor",
            (dtype,),
        )

    def forward(self):
        raise NotImplementedError(
            "SymmetricContractionBwdInfo module is not intended to be called."
        )

    def c_type(self):
        cg0 = self.make_clebsch_gordan_tensor(
            self.path_vals_w,
            self.path_indices_w,
            self.path_offsets_w,
            self.num_coupling_paths,
        )
        cg1 = self.make_clebsch_gordan_tensor(
            self.path_vals, self.path_indices, self.path_offsets, self.num_in_segments
        )
        return cg0, cg1


class _symmetric_tensor_contraction_autograd(torch.autograd.Function):
    @staticmethod
    def forward(ctx, in1_tensor, in2_tensor, in2_tensor_id, args, kernels):

        (num_out_segments, correlation, cg_fwd, cg0_bwd, cg1_bwd) = args
        (fwd_kernel, bwd_kernel) = kernels

        out_tensor = torch.zeros(
            [in1_tensor.shape[0], num_out_segments, in1_tensor.shape[-1]],
            dtype=in1_tensor.dtype,
            device=in1_tensor.device,
        )

        stream = torch.cuda.current_stream().cuda_stream

        fwd_kernel(
            maybe_detach(out_tensor),
            maybe_detach(in1_tensor),
            maybe_detach(in2_tensor),
            maybe_detach(in2_tensor_id),
            cg_fwd,
            correlation,
            stream_id=stream,
        )

        ctx.save_for_backward(in1_tensor, in2_tensor, in2_tensor_id)

        ctx.args = args
        ctx.kernels = kernels

        return out_tensor

    @staticmethod
    def backward(ctx, grad_out_tensor):

        if grad_out_tensor is None:
            return None, None, None, None, None

        (in1_tensor, in2_tensor, in2_tensor_id) = ctx.saved_tensors

        (_, correlation, _, cg0_bwd, cg1_bwd) = ctx.args
        (_, bwd_kernel) = ctx.kernels

        stream = torch.cuda.current_stream().cuda_stream

        d_in1_tensor = torch.zeros(
            in1_tensor.shape, dtype=in1_tensor.dtype, device=in1_tensor.device
        )

        accum_type = torch.float32 if in1_tensor.element_size() <= 4 else torch.float64

        d_in2_tensor = torch.zeros(
            in2_tensor.shape, dtype=accum_type, device=in2_tensor.device
        )

        bwd_kernel(
            maybe_detach(d_in1_tensor),
            maybe_detach(d_in2_tensor),
            maybe_detach(grad_out_tensor),
            maybe_detach(in1_tensor),
            maybe_detach(in2_tensor),
            maybe_detach(in2_tensor_id),
            cg0_bwd,
            cg1_bwd,
            correlation,
            stream,
        )

        if d_in2_tensor.dtype != in2_tensor.dtype:
            d_in2_tensor = d_in2_tensor.to(in2_tensor.dtype)

        return d_in1_tensor, d_in2_tensor, None, None, None


def _map_to_path_indices(path_segment_indices):
    path_indices = [[index[-1]] + index[0:-1] for index in path_segment_indices]
    return path_indices


class SymmetricTensorContraction(nn.Module):
    """
    A class to perform the symmetric tensor contraction in the Mace model

    Attributes
    ---------------------------
    num_in_segments : int
        number of segments in the fist input tensor
    num_couplings : int
        number of segments in the second input tensor
    num_out_segments : int
        number of segments in the output tensor
    correlation: int
        correlation length
    dtype: torch.dtype
        data type for computation
    fwd_info: SymmetricContractionFwdInfo
        object to store precompute data for the forward
    bwd_info: SymmetricContractionBwdInfo
        object to store precompute data for the backward

    Methods
    ---------------------------
    forward(in1_tensor, in2_tensor, in2_tensor_id):
        forward function for the symmetric tensor contraction in the Mace model
    get_segments_dims()
        get the numbers of the segments for the input and the output tensors
    """

    def __init__(
        self,
        path_segment_indices: list[list[int]],
        path_coefficients: list[float],
        num_in_segments: int,
        num_couplings: int,
        num_out_segments: int,
        correlation: int,
        dtype: torch.dtype,
    ):
        """
        Construct all the necessary attributes

        Parameters
        ---------------------------
        path_segment_indices: list[list[int]]
            list of integer list to represent each path
        path_coefficients: list[float]
            list of floating number for each path
        num_in_segments: int
            number of segments in the first input tensor,
            the repeated operand
        num_couplings: int
            number of segments in the second input tensor
        num_out_segments: int
            number of segments in the output tensor
        correlation: int
            correlation length for the symmetric tensor contraction
        dtype: torch.dtype
            data type to perform computation
        """
        super().__init__()

        if not torch.cuda.is_available():
            raise AssertionError("No Nvidia GPU is detected")

        if len(path_segment_indices) != len(path_coefficients):
            raise AssertionError(
                "Number of the path coefficients and of the path segment indices \
                have are different"
            )

        self.num_in_segments = num_in_segments
        self.num_couplings = num_couplings
        self.num_out_segments = num_out_segments

        self.correlation = correlation
        self.dtype = dtype

        path_indices = _map_to_path_indices(path_segment_indices)

        self.fwd_info = SymmetricContractionFwdInfo(
            path_coefficients, path_indices, self.num_out_segments, correlation, dtype
        )
        self.bwd_info = SymmetricContractionBwdInfo(
            path_coefficients,
            path_indices,
            self.num_in_segments,
            self.num_couplings,
            correlation,
            dtype,
        )

    def forward(self, in1_tensor, in2_tensor, in2_tensor_id):
        """
        Forward funciton call

        Parameters
        ---------------------------
        in1_tensor: torch.Tensor
            The input tensor with shape [num_batches, num_in_segments, num_features]
        in2_tensor: torch.Tensor
            The input tensor with shape [num_batches, num_couplings, num_features]
        in2_tensor_id:
            The ID with shape [num_batches]
        """
        cg_fwd = self.fwd_info.c_type()
        cg0_bwd, cg1_bwd = self.bwd_info.c_type()
        args = (
            self.num_out_segments,
            self.correlation,
            cg_fwd,
            cg0_bwd,
            cg1_bwd,
        )

        fwd_call = get_operator_from_module(
            ops,
            "symmetric_tensor_contraction_n2n_fwd",
            (self.dtype,),
        )
        bwd_call = get_operator_from_module(
            ops,
            "symmetric_tensor_contraction_n2n_bwd",
            (self.dtype,),
        )

        kernels = (fwd_call, bwd_call)

        return _symmetric_tensor_contraction_autograd.apply(
            in1_tensor, in2_tensor, in2_tensor_id, args, kernels
        )

    def get_segments_dims(self):
        """
        Get the numbers of the segments for the input and the output tensors
        """
        return (
            self.num_in_segments,
            self.num_out_segments,
            self.num_coupling_paths,
        )



# Copyright (c) 2023-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import torch
import torch.nn as nn
import pylibcugraphops

from dataclasses import dataclass
from typing import Tuple, Optional

from pylibcugraphops.pytorch.utils import (
    maybe_detach,
    maybe_empty_like,
    get_operator_from_module,
)
import pylibcugraphops.operators as ops


def sort_tp_info_for_operand(
    path_offsets_and_dims: torch.Tensor, path_cg_values: torch.Tensor, op_idx: int
):
    # idx = 0 -> in0
    # idx = 1 -> in1
    # idx = 2 -> in2
    # idx = 3 -> out
    key = path_offsets_and_dims[:, op_idx]
    perm = key.sort(descending=False, stable=True)[1]
    return path_offsets_and_dims[perm, :], path_cg_values[perm]


def path_offsets_to_csr(
    path_offsets_and_dims: torch.Tensor, op_idx: int, op_stride: int
):
    diff = torch.diff(
        path_offsets_and_dims[:, op_idx],
        n=1,
        append=path_offsets_and_dims.new_tensor([op_stride]),
    )
    path_csr_offsets = diff.nonzero().to(dtype=torch.int32).squeeze(dim=1)
    path_csr_offsets = torch.cat(
        [
            path_csr_offsets.new_tensor([0]),
            path_csr_offsets + 1,
        ]
    )
    return path_csr_offsets


def make_tp_info(
    op_idx: int,
    op_stride: int,
    path_offsets: torch.Tensor,
    path_cg_values: torch.Tensor,
):
    path_offsets = path_offsets.clone().detach()
    path_cg_values = path_cg_values.clone().detach()
    path_offsets, path_cg_values = sort_tp_info_for_operand(
        path_offsets, path_cg_values, op_idx
    )
    path_csr_offsets = path_offsets_to_csr(path_offsets, op_idx, op_stride)

    path_cg_values = path_cg_values
    path_offsets = path_offsets
    path_csr_offsets = path_csr_offsets

    return (
        path_csr_offsets,
        path_offsets,
        path_cg_values,
    )


def make_tp_infos(
    strides: list[int], path_offsets_and_dims: list[int], path_cg_values: list[float]
):
    path_cg_values = torch.FloatTensor(path_cg_values)
    path_offsets_and_dims = torch.IntTensor(path_offsets_and_dims)

    tp_infos = None

    if len(strides) == 4:
        tp_infos = (
            make_tp_info(3, strides[3], path_offsets_and_dims, path_cg_values),
            make_tp_info(0, strides[0], path_offsets_and_dims, path_cg_values),
            make_tp_info(1, strides[1], path_offsets_and_dims, path_cg_values),
            make_tp_info(2, strides[2], path_offsets_and_dims, path_cg_values),
        )

    elif len(strides) == 3:
        tp_infos = (
            make_tp_info(2, strides[2], path_offsets_and_dims, path_cg_values),
            make_tp_info(0, strides[0], path_offsets_and_dims, path_cg_values),
            make_tp_info(1, strides[1], path_offsets_and_dims, path_cg_values),
        )

    else:
        raise NotImplementedError

    return tp_infos


class TensorProductPathData(nn.Module):
    def __init__(
        self,
        path_csr_offsets: torch.Tensor,
        path_offsets: torch.Tensor,
        path_cg_values: torch.Tensor,
    ):
        super().__init__()
        # for now, don't have these buffers as part of the state dict to avoid
        # breaking changes w.r.t. loading state-dicts for now
        self.register_buffer("path_csr_offsets", path_csr_offsets, persistent=False)
        self.register_buffer("path_offsets", path_offsets, persistent=False)
        # Register float cg_values as int via bitcasting, so that model.to(dtype)
        # won't affect the precision of cg_values buffer
        self.register_buffer(
            "path_cg_values", path_cg_values.view(dtype=torch.int32), persistent=False
        )

    def forward(self, *args, **kwargs):
        raise NotImplementedError(
            "TensorProductPathData module is not intended to be called."
        )

    def c_type(self):
        return pylibcugraphops.make_tp_info(
            self.path_csr_offsets,
            self.path_offsets,
            self.path_cg_values.view(dtype=torch.float32),
        )


@dataclass
class TensorProductArgs:
    output_shape: Tuple[int, int]
    tp_info_fwd: pylibcugraphops.tp_info_fp32
    tp_info_bwd_dgrad_in0: pylibcugraphops.tp_info_fp32
    tp_info_bwd_dgrad_in1: pylibcugraphops.tp_info_fp32
    tp_info_bwd_dgrad_in2: pylibcugraphops.tp_info_fp32
    connection_mode: pylibcugraphops.ConnectionMode
    needs_input_grad_in0: bool
    needs_input_grad_in1: bool
    needs_input_grad_in2: bool
    needs_input_grad_bias: bool


class _fused_tensor_product_autograd(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        in0: torch.Tensor,
        in1: torch.Tensor,
        in2: Optional[torch.Tensor],
        bias: Optional[torch.Tensor],
        args: TensorProductArgs,
    ) -> torch.Tensor:

        out = torch.empty(args.output_shape, device=in0.device, dtype=in0.dtype)
        stream = torch.cuda.current_stream().cuda_stream

        fwd_fun = get_operator_from_module(
            ops,
            "fused_tp_n2n_fwd",
            (
                in0.dtype,
                in1.dtype,
                in2.dtype if in2 is not None else out.dtype,
                out.dtype,
            ),
        )

        fwd_fun(
            out,
            maybe_detach(in0),
            maybe_detach(in1),
            maybe_detach(in2),
            maybe_detach(bias),
            args.connection_mode,
            args.tp_info_fwd,
            stream_id=stream,
        )

        ctx.save_for_backward(
            in0,
            in1,
            in2,
            bias,
        )
        ctx.args = args

        return out

    @staticmethod
    def backward(
        ctx, grad_out: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, None]:

        (in0, in1, in2, bias) = ctx.saved_tensors

        return _fused_tensor_product_backward_autograd.apply(
            grad_out,
            in0,
            in1,
            in2,
            bias,
            ctx.args,
        )


class _fused_tensor_product_backward_autograd(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        grad_out: torch.Tensor,
        in0: torch.Tensor,
        in1: torch.Tensor,
        in2: Optional[torch.Tensor],
        bias: Optional[torch.Tensor],
        args: TensorProductArgs,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, None, None]:

        # calling function takes care of ensuring contiguous and detach
        grad_in0 = maybe_empty_like(in0, args.needs_input_grad_in0)
        grad_in1 = maybe_empty_like(in1, args.needs_input_grad_in1)
        grad_in2 = maybe_empty_like(in2, args.needs_input_grad_in2)
        grad_bias = maybe_empty_like(bias, args.needs_input_grad_bias)

        stream = torch.cuda.current_stream().cuda_stream

        bwd_fun = get_operator_from_module(
            ops,
            "fused_tp_n2n_bwd",
            (
                in0.dtype,
                in1.dtype,
                in2.dtype if in2 is not None else grad_out.dtype,
                grad_out.dtype,
            ),
        )

        in0 = maybe_detach(in0)
        in1 = maybe_detach(in1)
        in2 = maybe_detach(in2)
        bias = maybe_detach(bias)
        grad_out = maybe_detach(grad_out)

        bwd_fun(
            grad_in0,
            grad_in1,
            grad_in2,
            grad_bias,
            grad_out,
            in0,
            in1,
            in2,
            bias,
            args.connection_mode,
            args.tp_info_bwd_dgrad_in0,
            args.tp_info_bwd_dgrad_in1,
            args.tp_info_bwd_dgrad_in2,
            stream_id=stream,
        )

        ctx.args = args
        ctx.save_for_backward(
            in0,
            in1,
            in2,
            grad_out,
        )

        return grad_in0, grad_in1, grad_in2, grad_bias, None

    @staticmethod
    def backward(
        ctx, grad_grad_in0, grad_grad_in1, grad_grad_in2, grad_grad_bias, grad_args
    ):
        # args should not require_grad in any case
        # assert not ctx.needs_input_grad[4]
        in0, in1, in2, grad_out = ctx.saved_tensors

        grad_grad_out = maybe_empty_like(grad_out, ctx.needs_input_grad[0])
        grad_in0 = maybe_empty_like(in0, ctx.needs_input_grad[1])
        grad_in1 = maybe_empty_like(in1, ctx.needs_input_grad[2])
        grad_in2 = maybe_empty_like(in2, ctx.needs_input_grad[3])

        stream = torch.cuda.current_stream().cuda_stream

        bwd_bwd_fun = get_operator_from_module(
            ops,
            "fused_tp_n2n_bwd_bwd",
            (
                in0.dtype,
                in1.dtype,
                in2.dtype if in2 is not None else grad_out.dtype,
                grad_out.dtype,
            ),
        )

        bwd_bwd_fun(
            grad_in0,
            grad_in1,
            grad_in2,
            grad_grad_out,
            maybe_detach(grad_grad_in0),
            maybe_detach(grad_grad_in1),
            maybe_detach(grad_grad_in2),
            grad_out,
            in0,
            in1,
            in2,
            ctx.args.connection_mode,
            ctx.args.tp_info_fwd,
            ctx.args.tp_info_bwd_dgrad_in0,
            ctx.args.tp_info_bwd_dgrad_in1,
            ctx.args.tp_info_bwd_dgrad_in2,
            stream_id=stream,
        )

        if grad_grad_bias is not None:
            grad_grad_out += grad_grad_bias

        return grad_grad_out, grad_in0, grad_in1, grad_in2, None, None


class FusedTensorProductOp4(nn.Module):
    def __init__(
        self,
        operand_segment_modes: list[str],
        operand_segment_offsets: list[list[int]],
        operand_segment_shapes: list[list[list[int]]],
        path_indices: list[list[int]],
        path_coefficients: list[float],
    ):
        super().__init__()

        mode = "_".join(operand_segment_modes)

        mode_mappings = {
            "u__u_u": "UVU",
            "u__uw_w": "U1W",
            "u__uv_v": "U1W",
            "_v_vw_w": "1VW",
            "u_v_uv_u": "UVU",
            "u_v_uv_v": "UVV",
            "u_u_uw_w": "UUW",
            "u_v_uvw_w": "UVW",
        }

        # for know, we can always extract
        # u,v,w from individual operands
        dims_dict = {
            "u__u_u": (0, 1, -1),
            "u__uw_w": (0, 1, 3),
            "u__uv_v": (0, 1, 3),
            "_v_vw_w": (0, 1, 3),
            "u_v_uv_u": (0, 1, -1),
            "u_v_uv_v": (0, 1, -1),
            "u_u_uw_w": (0, -1, 3),
            "u_v_uvw_w": (0, 1, 3),
        }

        assert (
            mode in mode_mappings
        ), f"{mode} not found, expected {list(mode_mappings.keys())}"

        u_op_idx, v_op_idx, w_op_idx = dims_dict[mode]

        if (
            len(operand_segment_modes) != 4
            or len(operand_segment_offsets) != 4
            or len(operand_segment_shapes) != 4
        ):
            raise ValueError("Expected 4 operands (3 inputs, 1 output)")

        if len(path_indices) != len(path_coefficients):
            msg = (
                "path_coefficients and path_indices both have to be of length num_paths"
            )
            msg += f", got {len(path_coefficients)} & {len(path_indices)} respectively."

            raise ValueError(msg)

        operand_segment_strides = [
            [math.prod(segment) for segment in op] for op in operand_segment_shapes
        ]

        strides = [sum(op) for op in operand_segment_strides]

        path_offsets_and_dims = []
        self.connection_mode = mode_mappings[mode]

        for indices in path_indices:
            u = (
                operand_segment_strides[u_op_idx][indices[u_op_idx]]
                if u_op_idx != -1
                else -1
            )
            v = (
                operand_segment_strides[v_op_idx][indices[v_op_idx]]
                if v_op_idx != -1
                else -1
            )
            w = (
                operand_segment_strides[w_op_idx][indices[w_op_idx]]
                if w_op_idx != -1
                else -1
            )

            path_offsets_and_dims.append(
                (
                    operand_segment_offsets[0][indices[0]],
                    operand_segment_offsets[1][indices[1]],
                    operand_segment_offsets[2][indices[2]],
                    operand_segment_offsets[3][indices[3]],
                    u,
                    v,
                    w,
                    -1,
                )
            )

        tp_infos = make_tp_infos(strides, path_offsets_and_dims, path_coefficients)

        self.stride_out = strides[-1]
        self._tp_info_fwd = TensorProductPathData(*tp_infos[0])
        self._tp_info_bwd_dgrad_in0 = TensorProductPathData(*tp_infos[1])
        self._tp_info_bwd_dgrad_in1 = TensorProductPathData(*tp_infos[2])
        self._tp_info_bwd_dgrad_in2 = TensorProductPathData(*tp_infos[3])

    def forward(
        self,
        in0: torch.Tensor,
        in1: torch.Tensor,
        in2: torch.Tensor,
    ) -> torch.Tensor:
        if in0.dim() == 1 and in1.dim() == 1 and in2.dim() != 1:
            batch_size = in2.size(0)
        elif in0.dim() == 1 and in1.dim() != 1:
            batch_size = in1.size(0)
        else:
            batch_size = in0.size(0)

        args = TensorProductArgs(
            (batch_size, self.stride_out),
            self._tp_info_fwd.c_type(),
            self._tp_info_bwd_dgrad_in0.c_type(),
            self._tp_info_bwd_dgrad_in1.c_type(),
            self._tp_info_bwd_dgrad_in2.c_type(),
            getattr(pylibcugraphops.ConnectionMode, self.connection_mode),
            in0.requires_grad,
            in1.requires_grad,
            in2.requires_grad,
            False,
        )

        out = _fused_tensor_product_autograd.apply(in0, in1, in2, None, args)

        return out


class FusedTensorProductOp3(nn.Module):
    def __init__(
        self,
        operand_segment_modes: list[str],
        operand_segment_offsets: list[list[int]],
        operand_segment_shapes: list[list[list[int]]],
        path_indices: list[list[int]],
        path_coefficients: list[float],
    ):
        super().__init__()

        mode = "_".join(operand_segment_modes)

        mode_mappings = {
            "u_u_u": "UUU",
            "u_v_uv": "UVUV",
            "u_uv_v": "UUVV",
            "u__u": "U1U1",
            "_v_v": "1V1V",
        }

        # for know, we can always extract
        # u,v,w from individual operands
        dims_dict = {
            "u_u_u": (0, -1),
            "u_v_uv": (0, 1),
            "u_uv_v": (0, 2),
            "u__u": (0, 1),
            "_v_v": (0, 1),
        }

        assert mode in mode_mappings

        u_op_idx, v_op_idx = dims_dict[mode]

        if (
            len(operand_segment_modes) != 3
            or len(operand_segment_offsets) != 3
            or len(operand_segment_shapes) != 3
        ):
            raise ValueError("Expected 3 operands (3 inputs, 1 output)")

        if len(path_indices) != len(path_coefficients):
            msg = ""
            msg += (
                "path_coefficients and path_indices both have to be of length num_paths"
            )
            msg += f", got {len(path_coefficients)} & {len(path_indices)} respectively."
            raise ValueError(msg)

        operand_segment_strides = [
            [math.prod(segment) for segment in op] for op in operand_segment_shapes
        ]
        strides = [sum(op) for op in operand_segment_strides]

        path_offsets_and_dims = []
        self.connection_mode = mode_mappings[mode]

        for indices in path_indices:
            u = (
                operand_segment_strides[u_op_idx][indices[u_op_idx]]
                if u_op_idx != -1
                else -1
            )
            v = (
                operand_segment_strides[v_op_idx][indices[v_op_idx]]
                if v_op_idx != -1
                else -1
            )
            path_offsets_and_dims.append(
                (
                    operand_segment_offsets[0][indices[0]],
                    operand_segment_offsets[1][indices[1]],
                    -1,
                    operand_segment_offsets[2][indices[2]],
                    u,
                    v,
                    -1,
                    -1,
                )
            )

        tp_infos = make_tp_infos(strides, path_offsets_and_dims, path_coefficients)

        self.stride_out = strides[-1]
        self._tp_info_fwd = TensorProductPathData(*tp_infos[0])
        self._tp_info_bwd_dgrad_in0 = TensorProductPathData(*tp_infos[1])
        self._tp_info_bwd_dgrad_in1 = TensorProductPathData(*tp_infos[2])

    def forward(
        self,
        in0: torch.Tensor,
        in1: torch.Tensor,
    ) -> torch.Tensor:
        if in0.dim() == 1 and in1.dim() != 1:
            batch_size = in1.size(0)
        else:
            batch_size = in0.size(0)

        args = TensorProductArgs(
            (batch_size, self.stride_out),
            self._tp_info_fwd.c_type(),
            self._tp_info_bwd_dgrad_in0.c_type(),
            self._tp_info_bwd_dgrad_in1.c_type(),
            self._tp_info_bwd_dgrad_in1.c_type(),
            getattr(pylibcugraphops.ConnectionMode, self.connection_mode),
            in0.requires_grad,
            in1.requires_grad,
            False,
            False,
        )

        out = _fused_tensor_product_autograd.apply(in0, in1, None, None, args)

        return out



# Copyright (c) 2022-2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from ..utils import maybe_empty, get_tensor_meta_data
from ..graph import CSC
from torch.autograd.function import once_differentiable
from pylibcugraphops.operators import (
    agg_dmpnn_e2e_fwd,
    agg_dmpnn_e2e_bwd,
)


def agg_dmpnn_e2e(
    edge_emb: torch.Tensor,
    graph: CSC,
    rev_edge_ids: torch.Tensor,
    concat_own: bool,
) -> torch.Tensor:
    r"""PyTorch autograd function for the Directed Message Passing NN (agg_dmpnn)
    primitive which performs edge-based (e2e) aggregation taking the
    direction of edges into account.

    Parameters
    ----------
    edge_emb : torch.Tensor
        The input edge features.
        Shape: (n_edges, dim).

    graph : CSC
        The graph used for the operation.

    rev_edge_ids : torch.Tensor
        the position indices of each edge for the corresponding reverse edge
        Shape: (n_indices,).

    concat_own : bool
        flag indicating whether to concat the own edge features

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape: (n_edges, dim)

    """

    if graph.is_bipartite:
        raise ValueError("Operator does not support bipartite graphs.")

    return _agg_dmpnn_e2e_autograd.apply(edge_emb, graph, rev_edge_ids, concat_own)


class _agg_dmpnn_e2e_autograd(torch.autograd.Function):
    r"""Custom autograd function for the Directed Message Passing Neural Network (DMPNN)
    primitive which performs edge-based (e2e) aggregation taking the direction of edges
    into account while operating on a graph."""

    @staticmethod
    def forward(ctx, edge_emb, graph, rev_edge_ids, concat_own):
        if concat_own:
            out = torch.empty(
                (edge_emb.size(0), 2 * edge_emb.size(1)),
                dtype=edge_emb.dtype,
                device=edge_emb.device,
            )
        else:
            out = torch.empty_like(edge_emb)

        stream = torch.cuda.current_stream().cuda_stream
        edge_emb = edge_emb.detach().contiguous()
        agg_dmpnn_e2e_fwd(
            out,
            edge_emb,
            graph.graph_csc,
            rev_edge_ids,
            concat_own,
            stream_id=stream,
        )

        ctx.save_for_backward(rev_edge_ids)
        ctx.graph = graph
        ctx.concat_own = concat_own
        ctx.edge_emb_meta_data = get_tensor_meta_data(edge_emb)
        return out

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        (rev_edge_ids,) = ctx.saved_tensors
        needs_grad_emb, _, _, _ = ctx.needs_input_grad
        if not needs_grad_emb:
            return None, None, None, None

        grad_edge_emb = maybe_empty(**ctx.edge_emb_meta_data, create_tensor=True)

        stream = torch.cuda.current_stream().cuda_stream
        grad_output = grad_output.detach().contiguous()
        agg_dmpnn_e2e_bwd(
            grad_edge_emb,
            grad_output,
            ctx.graph.graph_csc,
            rev_edge_ids,
            ctx.concat_own,
            stream_id=stream,
        )
        return grad_edge_emb, None, None, None



# Copyright (c) 2022-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .agg_concat import agg_concat_n2n, agg_concat_e2n, agg_concat_n2n_e2n
from .agg_dmpnn import agg_dmpnn_e2e
from .agg_hg_basis import agg_hg_basis_n2n_post
from .agg_simple import agg_simple_e2n, agg_simple_n2n, agg_simple_n2n_e2n
from .mha_gat import mha_gat_n2n
from .mha_gat_v2 import mha_gat_v2_n2n
from .mha_simple import mha_simple_n2n
from .update_efeat import update_efeat_e2e
from .fused_tensor_product import FusedTensorProductOp3, FusedTensorProductOp4
from .fused_tensor_product_transpose import (
    transpose_segment_to_contiguous,
    transpose_segment_to_strided,
)
from .symmetric_tensor_contraction import (
    SymmetricTensorContraction,
)


__all__ = [
    "agg_concat_n2n",
    "agg_concat_e2n",
    "agg_concat_n2n_e2n",
    "agg_dmpnn_e2e",
    "agg_hg_basis_n2n_post",
    "agg_simple_n2n",
    "agg_simple_e2n",
    "agg_simple_n2n_e2n",
    "mha_gat_n2n",
    "mha_gat_v2_n2n",
    "mha_simple_n2n",
    "update_efeat_e2e",
    "FusedTensorProductOp3",
    "FusedTensorProductOp4",
    "transpose_irrep_to_m_last",
    "transpose_irrep_to_channels_last",
    "SymmetricTensorContraction",
]
